"""SIWave simulation setup."""

from ansys.edb.core.simulation_setup.simulation_setup import SimulationSetup, SimulationSetupType
from ansys.edb.core.simulation_setup.siwave_dcir_simulation_settings import (
    SIWaveDCIRSimulationSettings,
)


class SIWaveDCIRSimulationSetup(SimulationSetup):
    """Represents SIWave DCIR simulation setup data."""

    @classmethod
    def create(cls, cell, name):
        """Create a SIWave DCIR simulation setup.

        Parameters
        ----------
        cell : :class:`.Cell`
            Cell to create the simulation setup in.
        name : str
            Name of the simulation setup.

        Returns
        -------
        SIWaveDCIRSimulationSetup
            Simulation setup created.
        """
        return super()._create(cell, name, SimulationSetupType.SI_WAVE_DCIR)

    @property
    def settings(self):
        """:class:`.SIWaveSimulationSettings`: Simulation settings of the simulation setup."""
        return SIWaveDCIRSimulationSettings(self)



"""SIWave simulation settings."""

from enum import Enum

import ansys.api.edb.v1.si_wave_simulation_settings_pb2 as pb

from ansys.edb.core.inner import messages
from ansys.edb.core.session import (
    SIWaveAdvancedSettingsServiceStub,
    SIWaveDCAdvancedSettingsServiceStub,
    SIWaveDCSettingsServiceStub,
    SIWaveGeneralSettingsServiceStub,
    SIWaveSParameterSettingsServiceStub,
    StubAccessor,
    StubType,
)
from ansys.edb.core.simulation_setup.simulation_settings import (
    SimulationSettings,
    SimulationSettingsBase,
)


class SParamInterpolation(Enum):
    """Provides an enum representing s parameter interpolation types."""

    POINT_IN = pb.POINT_IN
    LINEAR_IN = pb.LINEAR_IN
    STEP_IN = pb.STEP_IN


class SParamExtrapolation(Enum):
    """Provides an enum representing s parameter extrapolation types."""

    ZERO_EX = pb.ZERO_EX
    SAME_EX = pb.SAME_EX
    LINEAR_EX = pb.LINEAR_EX
    CONSTANT_EX = pb.CONSTANT_EX


class SParamDCBehavior(Enum):
    """Provides an enum representing s parameter DC behavior types."""

    ZERO_DC = pb.ZERO_DC
    SAME_DC = pb.SAME_DC
    LINEAR_DC = pb.LINEAR_DC
    CONSTANT_DC = pb.CONSTANT_DC
    ONE_PORT_CAPACITOR_DC = pb.ONE_PORT_CAPACITOR_DC
    OPEN_DC = pb.OPEN_DC


class ACDCMergeMode(Enum):
    """Provides an enum representing AC/DC merge mode types."""

    DEFAULT = pb.DEFAULT
    FAST = pb.FAST


class SIWaveSimulationSettings(SimulationSettings):
    """Represents SIWave simulation settings."""

    @property
    def general(self):
        """:class:`.SIWaveGeneralSettings`: General settings for SIWave simulations."""
        return SIWaveGeneralSettings(self._sim_setup)

    @property
    def advanced(self):
        """:class:`.SIWaveAdvancedSettings`: Advanced settings for SIWave simulations."""
        return SIWaveAdvancedSettings(self._sim_setup)

    @property
    def dc(self):
        """:class:`.SIWaveDCSettings`: DC settings for SIWave simulations."""
        return SIWaveDCSettings(self._sim_setup)

    @property
    def dc_advanced(self):
        """:class:`.SIWaveDCAdvancedSettings`: Advanced DC settings for SIWave simulations."""
        return SIWaveDCAdvancedSettings(self._sim_setup)

    @property
    def s_parameter(self):
        """:class:`.SIWaveSParameterSettings`: S parameter settings for SIWave simulations."""
        return SIWaveSParameterSettings(self._sim_setup)


class SIWaveGeneralSettings(SimulationSettingsBase):
    """Represents general settings for SIWave simulations."""

    __stub: SIWaveGeneralSettingsServiceStub = StubAccessor(StubType.siwave_general_sim_settings)

    @property
    def use_si_settings(self):
        """:obj:`bool`: Flag indicating if SI or PI settings are used."""
        return self.__stub.GetUseSISettings(self.msg).value

    @use_si_settings.setter
    def use_si_settings(self, use_si_settings):
        self.__stub.SetUseSISettings(messages.bool_property_message(self, use_si_settings))

    @property
    def use_custom_settings(self):
        """:obj:`bool`: Flag indicating if custom settings are used rather than SI or PI settings."""
        return self.__stub.GetUseCustomSettings(self.msg).value

    @use_custom_settings.setter
    def use_custom_settings(self, use_si_settings):
        self.__stub.SetUseCustomSettings(messages.bool_property_message(self, use_si_settings))

    @property
    def si_slider_pos(self):
        """:obj:`int`: SI slider position value."""
        return self.__stub.GetSISliderPos(self.msg).value

    @si_slider_pos.setter
    def si_slider_pos(self, si_slider_pos):
        self.__stub.SetSISliderPos(messages.uint64_property_message(self, si_slider_pos))

    @property
    def pi_slider_pos(self):
        """:obj:`int`: PI slider position value."""
        return self.__stub.GetPISliderPos(self.msg).value

    @pi_slider_pos.setter
    def pi_slider_pos(self, pi_slider_pos):
        self.__stub.SetPISliderPos(messages.uint64_property_message(self, pi_slider_pos))


class SIWaveAdvancedSettings(SimulationSettingsBase):
    """Represents advanced settings for SIWave simulations."""

    __stub: SIWaveAdvancedSettingsServiceStub = StubAccessor(StubType.siwave_advanced_sim_settings)

    @property
    def include_co_plane_coupling(self):
        """:obj:`bool`: Flag indicating if the co-plane coupling is included."""
        return self.__stub.GetIncludeCoPlaneCoupling(self.msg).value

    @include_co_plane_coupling.setter
    def include_co_plane_coupling(self, include_co_plane_coupling):
        self.__stub.SetIncludeCoPlaneCoupling(
            messages.bool_property_message(self, include_co_plane_coupling)
        )

    @property
    def include_inter_plane_coupling(self):
        """:obj:`bool`: Flag indicating if the inter-plane coupling is included."""
        return self.__stub.GetIncludeInterPlaneCoupling(self.msg).value

    @include_inter_plane_coupling.setter
    def include_inter_plane_coupling(self, include_inter_plane_coupling):
        self.__stub.SetIncludeInterPlaneCoupling(
            messages.bool_property_message(self, include_inter_plane_coupling)
        )

    @property
    def include_split_plane_coupling(self):
        """:obj:`bool`: Flag indicating if the split-plane coupling is included."""
        return self.__stub.GetIncludeSplitPlaneCoupling(self.msg).value

    @include_split_plane_coupling.setter
    def include_split_plane_coupling(self, include_split_plane_coupling):
        self.__stub.SetIncludeSplitPlaneCoupling(
            messages.bool_property_message(self, include_split_plane_coupling)
        )

    @property
    def include_fringe_plane_coupling(self):
        """:obj:`bool`: Flag indicating if the fringe-plane coupling is included."""
        return self.__stub.GetIncludeFringePlaneCoupling(self.msg).value

    @include_fringe_plane_coupling.setter
    def include_fringe_plane_coupling(self, include_fringe_plane_coupling):
        self.__stub.SetIncludeFringePlaneCoupling(
            messages.bool_property_message(self, include_fringe_plane_coupling)
        )

    @property
    def include_trace_plane_coupling(self):
        """:obj:`bool`: Flag indicating if the trace-plane coupling is included."""
        return self.__stub.GetIncludeTracePlaneCoupling(self.msg).value

    @include_trace_plane_coupling.setter
    def include_trace_plane_coupling(self, include_trace_plane_coupling):
        self.__stub.SetIncludeTracePlaneCoupling(
            messages.bool_property_message(self, include_trace_plane_coupling)
        )

    @property
    def cross_talk_threshold(self):
        """:obj:`str`: Cross-talk threshold."""
        return self.__stub.GetCrossTalkThreshold(self.msg).value

    @cross_talk_threshold.setter
    def cross_talk_threshold(self, cross_talk_threshold):
        self.__stub.SetCrossTalkThreshold(
            messages.string_property_message(self, cross_talk_threshold)
        )

    @property
    def max_coupled_lines(self):
        """:obj:`int`: Maximum number of coupled lines."""
        return self.__stub.GetMaxCoupledLines(self.msg).value

    @max_coupled_lines.setter
    def max_coupled_lines(self, max_coupled_lines):
        self.__stub.SetMaxCoupledLines(messages.int_property_message(self, max_coupled_lines))

    @property
    def min_void_area(self):
        """:obj:`str`: Minimum void area."""
        return self.__stub.GetMinVoidArea(self.msg).value

    @min_void_area.setter
    def min_void_area(self, min_void_area):
        self.__stub.SetMinVoidArea(messages.string_property_message(self, min_void_area))

    @property
    def min_pad_area_to_mesh(self):
        """:obj:`str`: Minimum pad area to mesh."""
        return self.__stub.GetMinPadAreaToMesh(self.msg).value

    @min_pad_area_to_mesh.setter
    def min_pad_area_to_mesh(self, min_pad_area_to_mesh):
        self.__stub.SetMinPadAreaToMesh(
            messages.string_property_message(self, min_pad_area_to_mesh)
        )

    @property
    def min_plane_area_to_mesh(self):
        """:obj:`str`: Minimum plane area to mesh."""
        return self.__stub.GetMinPlaneAreaToMesh(self.msg).value

    @min_plane_area_to_mesh.setter
    def min_plane_area_to_mesh(self, min_plane_area_to_mesh):
        self.__stub.SetMinPlaneAreaToMesh(
            messages.string_property_message(self, min_plane_area_to_mesh)
        )

    @property
    def snap_length_threshold(self):
        """:obj:`str`: Snapping length threshold."""
        return self.__stub.GetSnapLengthThreshold(self.msg).value

    @snap_length_threshold.setter
    def snap_length_threshold(self, snap_length_threshold):
        self.__stub.SetSnapLengthThreshold(
            messages.string_property_message(self, snap_length_threshold)
        )

    @property
    def mesh_automatic(self):
        """:obj:`bool`: Flag indicating it mesh refinement frequency is automatically determined."""
        return self.__stub.GetMeshAutomatic(self.msg).value

    @mesh_automatic.setter
    def mesh_automatic(self, mesh_automatic):
        self.__stub.SetMeshAutomatic(messages.bool_property_message(self, mesh_automatic))

    @property
    def mesh_frequency(self):
        """:obj:`str`: Mesh refinement frequency."""
        return self.__stub.GetMeshFrequency(self.msg).value

    @mesh_frequency.setter
    def mesh_frequency(self, mesh_frequency):
        self.__stub.SetMeshFrequency(messages.string_property_message(self, mesh_frequency))

    @property
    def ac_dc_merge_mode(self):
        """:obj:`int`: AC/DC merge mode."""
        return ACDCMergeMode(self.__stub.GetAcDcMergeMode(self.msg).ac_dc_merge_mode)

    @ac_dc_merge_mode.setter
    def ac_dc_merge_mode(self, ac_dc_merge_mode):
        self.__stub.SetAcDcMergeMode(
            pb.ACDCMergeModePropertyMessage(
                target=self.msg, ac_dc_merge_mode=ac_dc_merge_mode.value
            )
        )

    @property
    def return_current_distribution(self):
        """:obj:`bool`: Flag indicating if return current distribution is traced."""
        return self.__stub.Get3DReturnCurrentDistribution(self.msg).value

    @return_current_distribution.setter
    def return_current_distribution(self, return_current_distribution):
        self.__stub.Set3DReturnCurrentDistribution(
            messages.bool_property_message(self, return_current_distribution)
        )

    @property
    def include_vi_sources(self):
        """:obj:`bool`: Flag indicating if voltage/current source connections/parasitics are included."""
        return self.__stub.GetIncludeVISources(self.msg).value

    @include_vi_sources.setter
    def include_vi_sources(self, include_vi_sources):
        self.__stub.SetIncludeVISources(messages.bool_property_message(self, include_vi_sources))

    @property
    def include_inf_gnd(self):
        """:obj:`bool`: Flag indicating if an infinite ground plane is included."""
        return self.__stub.GetIncludeInfGnd(self.msg).value

    @include_inf_gnd.setter
    def include_inf_gnd(self, include_inf_gnd):
        self.__stub.SetIncludeInfGnd(messages.bool_property_message(self, include_inf_gnd))

    @property
    def inf_gnd_location(self):
        """:obj:`str`: Infinite ground plane location."""
        return self.__stub.GetInfGndLocation(self.msg).value

    @inf_gnd_location.setter
    def inf_gnd_location(self, inf_gnd_location):
        self.__stub.SetInfGndLocation(messages.string_property_message(self, inf_gnd_location))

    @property
    def perform_erc(self):
        """:obj:`bool`: Flag indicating if ERC is performed during simulation setup."""
        return self.__stub.GetPerformERC(self.msg).value

    @perform_erc.setter
    def perform_erc(self, perform_erc):
        self.__stub.SetPerformERC(messages.bool_property_message(self, perform_erc))

    @property
    def ignore_non_functional_pads(self):
        """:obj:`bool`: Flag indicating if non-functional pads are ignored."""
        return self.__stub.GetIgnoreNonFunctionalPads(self.msg).value

    @ignore_non_functional_pads.setter
    def ignore_non_functional_pads(self, ignore_non_functional_pads):
        self.__stub.SetIgnoreNonFunctionalPads(
            messages.bool_property_message(self, ignore_non_functional_pads)
        )


class SIWaveDCSettings(SimulationSettingsBase):
    """Represents DC settings for SIWave simulations."""

    __stub: SIWaveDCSettingsServiceStub = StubAccessor(StubType.siwave_dc_sim_settings)

    @property
    def use_dc_custom_settings(self):
        """:obj:`bool`: Flag indicating if custom DC settings are used."""
        return self.__stub.GetUseDCCustomSettings(self.msg).value

    @use_dc_custom_settings.setter
    def use_dc_custom_settings(self, use_dc_custom_settings):
        self.__stub.SetUseDCCustomSettings(
            messages.bool_property_message(self, use_dc_custom_settings)
        )

    @property
    def compute_inductance(self):
        """:obj:`bool`: Flag indicating if inductance is calculated."""
        return self.__stub.GetComputeInductance(self.msg).value

    @compute_inductance.setter
    def compute_inductance(self, compute_inductance):
        self.__stub.SetComputeInductance(messages.bool_property_message(self, compute_inductance))

    @property
    def plot_jv(self):
        """:obj:`bool`: Flag indicating if current density and voltage distribution are plotted."""
        return self.__stub.GetPlotJV(self.msg).value

    @plot_jv.setter
    def plot_jv(self, plot_jv):
        self.__stub.SetPlotJV(messages.bool_property_message(self, plot_jv))

    @property
    def contact_radius(self):
        """:obj:`str`: Circuit element contact radius."""
        return self.__stub.GetContactRadius(self.msg).value

    @contact_radius.setter
    def contact_radius(self, contact_radius):
        self.__stub.SetContactRadius(messages.string_property_message(self, contact_radius))

    @property
    def dc_slider_pos(self):
        """:obj:`int`: DC slider position."""
        return self.__stub.GetDCSliderPos(self.msg).value

    @dc_slider_pos.setter
    def dc_slider_pos(self, dc_slider_pos):
        self.__stub.SetDCSliderPos(messages.uint64_property_message(self, dc_slider_pos))


class SIWaveDCAdvancedSettings(SimulationSettingsBase):
    """Represents advanced DC settings for SIWave simulations."""

    __stub: SIWaveDCAdvancedSettingsServiceStub = StubAccessor(
        StubType.siwave_dc_advanced_sim_settings
    )

    @property
    def dc_min_plane_area_to_mesh(self):
        """:obj:`str`: Geometry with an area smaller than this value is ignored."""
        return self.__stub.GetDCMinPlaneAreaToMesh(self.msg).value

    @dc_min_plane_area_to_mesh.setter
    def dc_min_plane_area_to_mesh(self, dc_min_plane_area_to_mesh):
        self.__stub.SetDCMinPlaneAreaToMesh(
            messages.string_property_message(self, dc_min_plane_area_to_mesh)
        )

    @property
    def dc_min_void_area_to_mesh(self):
        """:obj:`str`: Voids with an area smaller than this value are ignored."""
        return self.__stub.GetDCMinVoidAreaToMesh(self.msg).value

    @dc_min_void_area_to_mesh.setter
    def dc_min_void_area_to_mesh(self, dc_min_void_area_to_mesh):
        self.__stub.SetDCMinVoidAreaToMesh(
            messages.string_property_message(self, dc_min_void_area_to_mesh)
        )

    @property
    def max_init_mesh_edge_length(self):
        """:obj:`str`: Initial maximum edge length."""
        return self.__stub.GetMaxInitMeshEdgeLength(self.msg).value

    @max_init_mesh_edge_length.setter
    def max_init_mesh_edge_length(self, max_init_mesh_edge_length):
        self.__stub.SetMaxInitMeshEdgeLength(
            messages.string_property_message(self, max_init_mesh_edge_length)
        )

    @property
    def perform_adaptive_refinement(self):
        """:obj:`bool`: Flag indicating if adaptive refinement is performed."""
        return self.__stub.GetPerformAdaptiveRefinement(self.msg).value

    @perform_adaptive_refinement.setter
    def perform_adaptive_refinement(self, perform_adaptive_refinement):
        self.__stub.SetPerformAdaptiveRefinement(
            messages.bool_property_message(self, perform_adaptive_refinement)
        )

    @property
    def max_num_passes(self):
        """:obj:`int`: Maximum number of adaptive mesh refinement passes."""
        return self.__stub.GetMaxNumPasses(self.msg).value

    @max_num_passes.setter
    def max_num_passes(self, max_num_passes):
        self.__stub.SetMaxNumPasses(messages.uint64_property_message(self, max_num_passes))

    @property
    def min_num_passes(self):
        """:obj:`int`: Minimum number of adaptive mesh refinement passes."""
        return self.__stub.GetMinNumPasses(self.msg).value

    @min_num_passes.setter
    def min_num_passes(self, min_num_passes):
        self.__stub.SetMinNumPasses(messages.uint64_property_message(self, min_num_passes))

    @property
    def percent_local_refinement(self):
        """:obj:`int`: Percent of local refinement used for adaptive mesh refinement."""
        return self.__stub.GetPercentLocalRefinement(self.msg).value

    @percent_local_refinement.setter
    def percent_local_refinement(self, percent_local_refinement):
        self.__stub.SetPercentLocalRefinement(
            messages.uint64_property_message(self, percent_local_refinement)
        )

    @property
    def energy_error(self):
        """:obj:`float`: Percent of energy error used for adaptive mesh refinement."""
        return self.__stub.GetEnergyError(self.msg).value

    @energy_error.setter
    def energy_error(self, energy_error):
        self.__stub.SetEnergyError(messages.double_property_message(self, energy_error))

    @property
    def mesh_bws(self):
        """:obj:`bool`: Flag indicating if bondwires are meshed."""
        return self.__stub.GetMeshBws(self.msg).value

    @mesh_bws.setter
    def mesh_bws(self, mesh_bws):
        self.__stub.SetMeshBws(messages.bool_property_message(self, mesh_bws))

    @property
    def refine_bws(self):
        """:obj:`bool`: Flag indicating if the mesh along bondwires is refined."""
        return self.__stub.GetRefineBws(self.msg).value

    @refine_bws.setter
    def refine_bws(self, refine_bws):
        self.__stub.SetRefineBws(messages.bool_property_message(self, refine_bws))

    @property
    def mesh_vias(self):
        """:obj:`bool`: Flag indicating if vias are meshed."""
        return self.__stub.GetMeshVias(self.msg).value

    @mesh_vias.setter
    def mesh_vias(self, mesh_vias):
        self.__stub.SetMeshVias(messages.bool_property_message(self, mesh_vias))

    @property
    def refine_vias(self):
        """:obj:`bool`: Flag indicating if the mesh along vias is refined."""
        return self.__stub.GetRefineVias(self.msg).value

    @refine_vias.setter
    def refine_vias(self, refine_vias):
        self.__stub.SetRefineVias(messages.bool_property_message(self, refine_vias))

    @property
    def num_bw_sides(self):
        """:obj:`int`: Number of sides to use to approximate cylindrical bondwires."""
        return self.__stub.GetNumBwSides(self.msg).value

    @num_bw_sides.setter
    def num_bw_sides(self, num_bw_sides):
        self.__stub.SetNumBwSides(messages.uint64_property_message(self, num_bw_sides))

    @property
    def num_via_sides(self):
        """:obj:`int`: Number of sides to use to approximate cylindrical vias."""
        return self.__stub.GetNumViaSides(self.msg).value

    @num_via_sides.setter
    def num_via_sides(self, num_via_sides):
        self.__stub.SetNumViaSides(messages.uint64_property_message(self, num_via_sides))


class SIWaveSParameterSettings(SimulationSettingsBase):
    """Represents s parameter settings for SIWave simulations."""

    __stub: SIWaveSParameterSettingsServiceStub = StubAccessor(StubType.siwave_s_param_sim_settings)

    @property
    def use_state_space(self):
        """:obj:`bool`: Flag indicating if state space is used. If ``False``, a custom model is used."""
        return self.__stub.GetUseStateSpace(self.msg).value

    @use_state_space.setter
    def use_state_space(self, use_state_space):
        self.__stub.SetUseStateSpace(messages.bool_property_message(self, use_state_space))

    @property
    def interpolation(self):
        """:obj:`SParamInterpolation`: Interpolation type."""
        return SParamInterpolation(self.__stub.GetInterpolation(self.msg).interpolation)

    @interpolation.setter
    def interpolation(self, interpolation):
        self.__stub.SetInterpolation(
            pb.SParamInterpolationPropertyMessage(
                target=self.msg, interpolation=interpolation.value
            )
        )

    @property
    def extrapolation(self):
        """:obj:`SParamExtrapolation`: Extrapolation type."""
        return SParamExtrapolation(self.__stub.GetExtrapolation(self.msg).extrapolation)

    @extrapolation.setter
    def extrapolation(self, extrapolation):
        self.__stub.SetExtrapolation(
            pb.SParamExtrapolationPropertyMessage(
                target=self.msg, extrapolation=extrapolation.value
            )
        )

    @property
    def dc_behavior(self):
        """:obj:`SParamDCBehavior`: DC behavior type."""
        return SParamDCBehavior(self.__stub.GetDCBehavior(self.msg).dc_behavior)

    @dc_behavior.setter
    def dc_behavior(self, dc_behavior):
        self.__stub.SetDCBehavior(
            pb.SParamDCBehaviorPropertyMessage(target=self.msg, dc_behavior=dc_behavior.value)
        )



"""Layer map."""
from __future__ import annotations

from enum import Enum
from typing import List

from ansys.api.edb.v1 import layer_map_pb2 as pb

from ansys.edb.core.inner import ObjBase, messages
from ansys.edb.core.session import StubAccessor, StubType


class LayerMapUniqueDirection(Enum):
    """Provides an enum representing a unique direction."""

    FORWARD_UNIQUE = pb.FORWARD_UNIQUE
    """FORWARD_UNIQUE Mapping many to one (1,1), (2,1), (3,1)"""
    BACKWARD_UNIQUE = pb.BACKWARD_UNIQUE
    """BACKWARD_UNIQUE Mapping one to many (1,1), (1,2), (1,3)"""
    TWOWAY_UNIQUE = pb.TWOWAY_UNIQUE
    """TWOWAY_UNIQUE Mapping one to one (1,1), (2,3), (3,2)"""


class LayerMap(ObjBase):
    """Represents a two-way layer map where the key and value is the layer ID."""

    __stub = StubAccessor(StubType.layer_map)

    @staticmethod
    def create(direction: LayerMapUniqueDirection) -> LayerMap:
        """Create a layer map object.

        Parameters
        ----------
        direction : .LayerMapUniqueDirection
                Variable representing the map's direction.

        Returns
        -------
        .LayerMap
            LayerMap object created.
        """
        return LayerMap(
            StubAccessor(StubType.layer_map)
            .__get__()
            .Create(pb.LayerMapUniqueDirectionMessage(direction=direction.value))
        )

    def clear(self):
        """Clear the entrties of the layer map."""
        self.__stub.Clear(
            self.msg,
        )

    def set_mapping(self, from_id: int, to_id: int):
        """Create an entry in the layer map object.

        Parameters
        ----------
        from_id : int
            Layer ID (key) "from" which to map with the ``to_id`` parameter.
        to_id : int
            Layer ID (value) "to" which to map with the ``from_id`` parameter.
        """
        self.__stub.SetMapping(
            pb.LayerMapTwoIntPropertiesMessage(
                target=self.msg,
                from_id=messages.edb_internal_id_message(from_id),
                to_id=messages.edb_internal_id_message(to_id),
            )
        )

    def get_mapping_forward(self, layer_id: int) -> List[int]:
        """Get the list of IDs mapped forward with the given ID (key).

        Parameters
        ----------
        layer_id : int

        Returns
        -------
        list of int
        """
        msg = self.__stub.GetMappingForward(
            messages.int_property_message(
                target=self,
                value=layer_id,
            )
        )
        return [int(to_id) for to_id in msg.ids]

    def get_mapping_backward(self, layer_id: int) -> List[int]:
        """Get a list of IDs mapped backward with the given ID (value).

        Parameters
        ----------
        layer_id : int

        Returns
        -------
        list of int
        """
        msg = self.__stub.GetMappingBackward(
            messages.int_property_message(
                target=self,
                value=layer_id,
            )
        )
        return [int(from_id) for from_id in msg.ids]



"""HFSSPI simulation settings."""
from __future__ import annotations

from enum import Enum
from typing import List

import ansys.api.edb.v1.hfss_pi_simulation_settings_pb2 as pb

from ansys.edb.core.inner import messages
from ansys.edb.core.session import (
    HFSSPIGeneralSettingsServiceStub,
    HFSSPINetProcessingSettingsServiceStub,
    HFSSPIPowerGroundNetsServiceStub,
    HFSSPISignalNetsSettingsServiceStub,
    StubAccessor,
    StubType,
)
from ansys.edb.core.simulation_setup.simulation_settings import (
    SimulationSettings,
    SimulationSettingsBase,
)


class HFSSPISimulationPreference(Enum):
    """Provides an enum representing HFSSPI simulation preferences."""

    BALANCED = pb.BALANCED
    ACCURACY = pb.ACCURACY


class HFSSPIModelType(Enum):
    """Provides an enum representing HFSSPI model types."""

    RDL = pb.RDL
    PACKAGE = pb.PACKAGE
    PCB = pb.PCB


class SurfaceRoughnessModel(Enum):
    """Provides an enum representing HFSSPI surface roughness models."""

    NONE = pb.NONE
    EXPONENTIAL = pb.EXPONENTIAL
    HAMMERSTAD = pb.HAMMERSTAD


class ImprovedLossModel(Enum):
    """Provides an enum representing HFSSPI improved loss models."""

    LEVEL_1 = pb.LEVEL_1
    LEVEL_2 = pb.LEVEL_2
    LEVEL_3 = pb.LEVEL_3


class ConductorModeling(Enum):
    """Provides an enum representing HFSSPI conductor modeling options."""

    MESH_INSIDE = pb.MESH_INSIDE
    IMPEDANCE_BOUNDARY = pb.IMPEDANCE_BOUNDARY


class ErrorTolerance(Enum):
    """Provides an enum representing HFSSPI error tolerance values."""

    ET_0_0 = pb.ET_0_0
    ET_0_02 = pb.ET_0_02
    ET_0_04 = pb.ET_0_04
    ET_0_06 = pb.ET_0_06
    ET_0_08 = pb.ET_0_08
    ET_0_1 = pb.ET_0_1
    ET_0_2 = pb.ET_0_2
    ET_0_5 = pb.ET_0_5
    ET_1_0 = pb.ET_1_0


class HFSSPISimulationSettings(SimulationSettings):
    """Represents HFSSPI simulation settings."""

    @property
    def general(self) -> HFSSPIGeneralSettings:
        """:class:`.HFSSPIGeneralSettings`: General settings for HFSSPI simulations."""
        return HFSSPIGeneralSettings(self)

    @property
    def net_processing(self) -> HFSSPINetProcessingSettings:
        """:class:`.HFSSPINetProcessingSettings`: Net processing settings for HFSSPI simulations."""
        return HFSSPINetProcessingSettings(self)

    @property
    def power_ground_nets(self) -> HFSSPIPowerGroundNetsSettings:
        """:class:`.HFSSPIPowerGroundNetsSettings`: Power/ground nets settings for HFSSPI simulations."""
        return HFSSPIPowerGroundNetsSettings(self)

    @property
    def signal_nets(self) -> HFSSPISignalNetsSettings:
        """:class:`.HFSSPISignalNetsSettings`: Signal nets settings for HFSSPI simulations."""
        return HFSSPISignalNetsSettings(self)


class HFSSPIGeneralSettings(SimulationSettingsBase):
    """Represents HFSSPI general settings."""

    __stub: HFSSPIGeneralSettingsServiceStub = StubAccessor(StubType.hfss_pi_general_sim_settings)

    @property
    def simulation_preference(self) -> HFSSPISimulationPreference:
        """:class:`.HFSSPISimulationPreference`: Simulation preference. \
            Balanced (i.e., use less memory) or Accuracy (i.e., use more memory)."""
        return HFSSPISimulationPreference(
            self.__stub.GetPISliderPos(self.msg).hfss_pi_simulation_preference
        )

    @simulation_preference.setter
    def simulation_preference(self, simulation_preference: HFSSPISimulationPreference):
        self.__stub.SetPISliderPos(
            pb.HFSSPISimulationPreferencePropertyMessage(
                target=self.msg, hfss_pi_simulation_preference=simulation_preference.value
            )
        )

    @property
    def model_type(self) -> HFSSPIModelType:
        """:class:`.HFSSPIModelType`: (General Mode Only) Model type."""
        return HFSSPIModelType(self.__stub.GetHFSSPIModelType(self.msg).hfss_pi_model_type)

    @model_type.setter
    def model_type(self, hfss_pi_model_type: HFSSPIModelType):
        self.__stub.SetHFSSPIModelType(
            pb.HFSSPIModelTypePropertyMessage(
                target=self.msg, hfss_pi_model_type=hfss_pi_model_type.value
            )
        )

    @property
    def min_plane_area_to_mesh(self) -> str:
        """:obj:`str`: (General Mode Only) The minimum plane area to mesh. \
            Ignore geometry smaller than this value."""
        return self.__stub.GetMinPlaneAreaToMesh(self.msg).value

    @min_plane_area_to_mesh.setter
    def min_plane_area_to_mesh(self, min_plane_area_to_mesh: str):
        self.__stub.SetMinPlaneAreaToMesh(
            messages.string_property_message(self, min_plane_area_to_mesh)
        )

    @property
    def min_void_area_to_mesh(self) -> str:
        """:obj:`str`: (General Mode Only) The minimum void area to mesh. Ignore voids smaller than this value."""
        return self.__stub.GetMinVoidAreaToMesh(self.msg).value

    @min_void_area_to_mesh.setter
    def min_void_area_to_mesh(self, min_void_area_to_mesh: str):
        self.__stub.SetMinVoidAreaToMesh(
            messages.string_property_message(self, min_void_area_to_mesh)
        )

    @property
    def snap_length_threshold(self) -> str:
        """:obj:`str`: (General Mode Only) The snap length threshold. \
            Snap vertices separated by less than this value."""
        return self.__stub.GetSnapLengthThreshold(self.msg).value

    @snap_length_threshold.setter
    def snap_length_threshold(self, snap_length_threshold: str):
        self.__stub.SetSnapLengthThreshold(
            messages.string_property_message(self, snap_length_threshold)
        )

    @property
    def include_enhanced_bondwire_modeling(self) -> bool:
        """:obj:`bool`: Flag indicating whether to include enhanced bondwire modeling. This can slow simulation."""
        return self.__stub.GetIncludeEnhancedBondWireModeling(self.msg).value

    @include_enhanced_bondwire_modeling.setter
    def include_enhanced_bondwire_modeling(self, include_enhanced_bondwire_modeling: bool):
        self.__stub.SetIncludeEnhancedBondWireModeling(
            messages.bool_property_message(self, include_enhanced_bondwire_modeling)
        )

    @property
    def surface_roughness_model(self) -> SurfaceRoughnessModel:
        """:class:`.SurfaceRoughnessModel`: Surface roughness model."""
        return SurfaceRoughnessModel(
            self.__stub.GetSurfaceRoughnessModel(self.msg).surface_roughness_model
        )

    @surface_roughness_model.setter
    def surface_roughness_model(self, surface_roughness_model: SurfaceRoughnessModel):
        self.__stub.SetSurfaceRoughnessModel(
            pb.SurfaceRoughnessModelPropertyMessage(
                target=self.msg, surface_roughness_model=surface_roughness_model.value
            )
        )

    @property
    def rms_surface_roughness(self) -> str:
        """:obj:`str`: RMS surface roughness."""
        return self.__stub.GetRMSSurfaceRoughness(self.msg).value

    @rms_surface_roughness.setter
    def rms_surface_roughness(self, rms_surface_roughness: str):
        self.__stub.SetRMSSurfaceRoughness(
            messages.string_property_message(self, rms_surface_roughness)
        )

    @property
    def perform_erc(self) -> bool:
        """:obj:`bool`: (General Mode Only) Flag indicating whether to \
            perform error checking while generating the solver input file."""
        return self.__stub.GetPerformERC(self.msg).value

    @perform_erc.setter
    def perform_erc(self, perform_erc: bool):
        self.__stub.SetPerformERC(messages.bool_property_message(self, perform_erc))


class HFSSPINetProcessingSettings(SimulationSettingsBase):
    """Represents HFSSPI net processing settings."""

    __stub: HFSSPINetProcessingSettingsServiceStub = StubAccessor(
        StubType.hfss_pi_net_processing_sim_settings
    )

    @property
    def auto_select_nets_for_simulation(self) -> bool:
        """:obj:`bool`: Flag indicating whether to automatically select nets for simulation."""
        return self.__stub.GetAutoSelectNetsForSimulation(self.msg).value

    @auto_select_nets_for_simulation.setter
    def auto_select_nets_for_simulation(self, auto_select_nets_for_simulation: bool):
        self.__stub.SetAutoSelectNetsForSimulation(
            messages.bool_property_message(self, auto_select_nets_for_simulation)
        )

    @property
    def ignore_dummy_nets_for_selected_nets(self) -> bool:
        """:obj:`bool`: Flag indicating whether to ignore dummy nets for selected nets."""
        return self.__stub.GetIgnoreDummyNetsForSelectedNets(self.msg).value

    @ignore_dummy_nets_for_selected_nets.setter
    def ignore_dummy_nets_for_selected_nets(self, ignore_dummy_nets_for_selected_nets: bool):
        self.__stub.SetIgnoreDummyNetsForSelectedNets(
            messages.bool_property_message(self, ignore_dummy_nets_for_selected_nets)
        )

    @property
    def include_nets(self) -> List[str]:
        """:obj:`list` of :obj:`str`: Nets to include in HFSSPI simulation."""
        return self.__stub.GetIncludeNets(self.msg).strings

    @include_nets.setter
    def include_nets(self, value: List[str]):
        self.__stub.SetIncludeNets(messages.strings_property_message(self, value))


class HFSSPIPowerGroundNetsSettings(SimulationSettingsBase):
    """Represents HFSSPI power/ground nets settings."""

    __stub: HFSSPIPowerGroundNetsServiceStub = StubAccessor(
        StubType.hfss_pi_power_ground_sim_settings
    )

    @property
    def improved_loss_model(self):
        """:class:`.ImprovedLossModel`: Improved loss model."""
        return ImprovedLossModel(self.__stub.GetImprovedLossModel(self.msg).improved_loss_model)

    @improved_loss_model.setter
    def improved_loss_model(self, improved_loss_model: ImprovedLossModel):
        self.__stub.SetImprovedLossModel(
            pb.ImprovedLossModelPropertyMessage(
                target=self.msg, improved_loss_model=improved_loss_model.value
            )
        )

    @property
    def auto_detect_ignore_small_holes_min_diameter(self) -> bool:
        """:obj:`bool`: Flag indicating whether to automatically detect a diameter \
            that holes smaller than the diameter will be ignored."""
        return self.__stub.GetAutoDetectIgnoreSmallHolesMinDiameter(self.msg).value

    @auto_detect_ignore_small_holes_min_diameter.setter
    def auto_detect_ignore_small_holes_min_diameter(
        self, auto_detect_ignore_small_holes_min_diameter: bool
    ):
        self.__stub.SetAutoDetectIgnoreSmallHolesMinDiameter(
            messages.bool_property_message(self, auto_detect_ignore_small_holes_min_diameter)
        )

    @property
    def ignore_small_holes_min_diameter(self) -> str:
        """:obj:`str`: Diameter that holes smaller than the diameter will be ignored."""
        return self.__stub.GetIgnoreSmallHolesMinDiameter(self.msg).value

    @ignore_small_holes_min_diameter.setter
    def ignore_small_holes_min_diameter(self, ignore_small_holes_min_diameter: str):
        self.__stub.SetIgnoreSmallHolesMinDiameter(
            messages.string_property_message(self, ignore_small_holes_min_diameter)
        )


class HFSSPISignalNetsSettings(SimulationSettingsBase):
    """Represents HFSSPI signal nets settings."""

    __stub: HFSSPISignalNetsSettingsServiceStub = StubAccessor(
        StubType.hfss_pi_signal_nets_sim_settings
    )

    @property
    def error_tolerance(self) -> ErrorTolerance:
        """:class:`.ErrorTolerance`: Error tolerance."""
        return ErrorTolerance(self.__stub.GetSignalNetsErrorTolerance(self.msg).error_tolerance)

    @error_tolerance.setter
    def error_tolerance(self, error_tolerance: ErrorTolerance):
        self.__stub.SetSignalNetsErrorTolerance(
            pb.ErrorTolerancePropertyMessage(target=self.msg, error_tolerance=error_tolerance.value)
        )

    @property
    def conductor_modeling(self) -> ConductorModeling:
        """:class:`.ConductorModeling`: Conductor modeling. \
            When using surface roughness, users must use IMPEDANCE_BOUNDARY."""
        return ConductorModeling(
            self.__stub.GetSignalNetsConductorModeling(self.msg).conductor_modeling
        )

    @conductor_modeling.setter
    def conductor_modeling(self, conductor_modeling: ConductorModeling):
        self.__stub.SetSignalNetsConductorModeling(
            pb.ConductorModelingPropertyMessage(
                target=self.msg, conductor_modeling=conductor_modeling.value
            )
        )

    @property
    def include_improved_loss_handling(self) -> bool:
        """:obj:`bool`: Flag indicating whether to include improved metal loss handling. \
            Enabling this option can significantly slow simulation time."""
        return self.__stub.GetSignalNetsIncludeImprovedLossHandling(self.msg).value

    @include_improved_loss_handling.setter
    def include_improved_loss_handling(self, include_improved_loss_handling: bool):
        self.__stub.SetSignalNetsIncludeImprovedLossHandling(
            messages.bool_property_message(self, include_improved_loss_handling)
        )

    @property
    def include_improved_dielectric_fill_refinement(self) -> bool:
        """:obj:`bool`: Flag indicating whether to include improved dielectric fill refinement. \
            Activating Improved Dielectric Fill Refinement in Metal Layers can significantly slow simulation time."""
        return self.__stub.GetSignalNetsIncludeImprovedDielectricFillRefinement(self.msg).value

    @include_improved_dielectric_fill_refinement.setter
    def include_improved_dielectric_fill_refinement(
        self, include_improved_dielectric_fill_refinement: bool
    ):
        self.__stub.SetSignalNetsIncludeImprovedDielectricFillRefinement(
            messages.bool_property_message(self, include_improved_dielectric_fill_refinement)
        )



"""Component group."""

from enum import Enum

from ansys.api.edb.v1.component_group_pb2_grpc import ComponentGroupServiceStub
import ansys.api.edb.v1.edb_defs_pb2 as edb_defs_pb2

from ansys.edb.core.definition import component_property
from ansys.edb.core.hierarchy.group import Group
from ansys.edb.core.inner import messages
from ansys.edb.core.session import StubAccessor, StubType


class ComponentType(Enum):
    """Provides an enum representing component types."""

    OTHER = edb_defs_pb2.OTHER
    RESISTOR = edb_defs_pb2.RESISTOR
    INDUCTOR = edb_defs_pb2.INDUCTOR
    CAPACITOR = edb_defs_pb2.CAPACITOR
    IC = edb_defs_pb2.IC
    IO = edb_defs_pb2.IO
    INVALID = edb_defs_pb2.INVALID


class ComponentGroup(Group):
    """Represents a component group object."""

    __stub: ComponentGroupServiceStub = StubAccessor(StubType.component_group)

    @classmethod
    def create(cls, layout, name, comp_name):
        """Create a component group with a component.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the component group in.
        name : str
            Name of the component group.
        comp_name : str
            Name of the :class:`.ComponentDef` instance that \
            the component group refers to.

        Returns
        -------
        ComponentGroup
            Component group created.
        """
        return ComponentGroup(
            cls.__stub.Create(messages.component_group_create_message(layout, name, comp_name))
        )

    @property
    def num_pins(self):
        """:obj:`int`: Number of pins in the component group.

        This property is read-only.
        """
        return self.__stub.GetNumberOfPins(self.msg).value

    @property
    def component_property(self):
        """:obj:`ComponentProperty`: Component property of the component group.

        A copy is returned. Use the setter for any modifications to be reflected.
        """
        from ansys.edb.core.definition.ic_component_property import ICComponentProperty
        from ansys.edb.core.definition.io_component_property import IOComponentProperty
        from ansys.edb.core.definition.rlc_component_property import RLCComponentProperty

        comp_prop = component_property.ComponentProperty(self.__stub.GetComponentProperty(self.msg))
        comp_type = self.__stub.GetComponentType(self.msg).comp_type
        if (
            comp_type == edb_defs_pb2.ComponentType.RESISTOR
            or comp_type == edb_defs_pb2.ComponentType.INDUCTOR
            or comp_type == edb_defs_pb2.ComponentType.CAPACITOR
        ):
            return RLCComponentProperty(comp_prop.msg)
        if comp_type == edb_defs_pb2.ComponentType.IC:
            return ICComponentProperty(comp_prop.msg)
        if (
            comp_type == edb_defs_pb2.ComponentType.IO
            or comp_type == edb_defs_pb2.ComponentType.OTHER
        ):
            return IOComponentProperty(comp_prop.msg)
        else:
            return None

    @component_property.setter
    def component_property(self, value):
        """Component property on the component group."""
        self.__stub.SetComponentProperty(messages.pointer_property_message(self, value))

    @property
    def component_type(self):
        """:obj:`ComponentType`: Component type of the component group."""
        return ComponentType(self.__stub.GetComponentType(self.msg).comp_type)

    @component_type.setter
    def component_type(self, value):
        """Component type on the component group."""
        self.__stub.SetComponentType(messages.set_component_group_type_message(self, value))

    @classmethod
    def find_by_def(cls, layout, comp_def_name):
        """Find all components belonging to a given component definition.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to search for component groups.
        comp_def_name : str
            Name of the :class:`.ComponentDef` instance.

        Returns
        -------
        list[ComponentGroup]
            List of component groups that are found.
        """
        objs = cls.__stub.FindByDef(
            messages.object_name_in_layout_message(layout, comp_def_name)
        ).items
        return [ComponentGroup(cg) for cg in objs]



"""Bundle Terminal."""

from ansys.edb.core.inner import TypeField, messages
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.terminal.terminal import Terminal, TerminalType


class BundleTerminal(Terminal):
    """Represents a bundle terminal object."""

    __stub = StubAccessor(StubType.bundle_terminal)
    type = TypeField(TerminalType.BUNDLE)

    @classmethod
    def create(cls, terminals):
        """Create a bundle terminal.

        Parameters
        ----------
        terminals : list of Terminal

        Returns
        -------
        BundleTerminal
        """
        return BundleTerminal(cls.__stub.Create(messages.edb_obj_collection_message(terminals)))

    @property
    def terminals(self):
        """:obj:`list` of Terminal: All terminals grouped in the terminal."""
        return [Terminal(msg).cast() for msg in self.__stub.GetTerminals(self.msg).items]

    def ungroup(self):
        """Delete the grouping."""
        self.__stub.Ungroup(self.msg)
        self.msg = None



"""Pin group."""

from ansys.edb.core.edb_defs import LayoutObjType
from ansys.edb.core.inner import messages
from ansys.edb.core.inner.conn_obj import ConnObj
from ansys.edb.core.primitive.padstack_instance import PadstackInstance
from ansys.edb.core.session import PinGroupServiceStub, StubAccessor, StubType
from ansys.edb.core.terminal.pin_group_terminal import PinGroupTerminal


class PinGroup(ConnObj):
    """Represents a pin group object."""

    __stub: PinGroupServiceStub = StubAccessor(StubType.pin_group)
    layout_obj_type = LayoutObjType.PIN_GROUP

    @classmethod
    def create(cls, layout, name, padstack_instances):
        """Create a pin group.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the pin group in.
        name : str
            Name of the pin group.
        padstack_instances : list[:class:`.PadstackInstance`]
            List of padstack instances.

        Returns
        -------
        PinGroup
            Pin group created.
        """
        return PinGroup(
            cls.__stub.Create(messages.pin_group_creation_message(layout, name, padstack_instances))
        )

    @classmethod
    def find(cls, layout, name):
        """Find a pin group by name in a given layout.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to search for the pin group.
        name : str
            Name of the pin group.

        Returns
        -------
        PinGroup
            Pin group found, ``None`` otherwise.
        """
        return PinGroup(cls.__stub.FindByName(messages.pin_group_lookup_message(layout, name)))

    @classmethod
    def unique_name(cls, layout, prefix):
        """Get a unique pin group name in the layout using a given prefix.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to search for the pin group.
        prefix : str
            Prefix of the unique name.

        Returns
        -------
        str
            Name of the pin group found.
        """
        return cls.__stub.GetUniqueName(
            messages.pin_group_get_unique_name_message(layout, prefix)
        ).value

    @property
    def name(self):
        """:obj:`str`: Name of the pin group.

        This property is read-only.
        """
        return self.__stub.GetName(self.msg).value

    @property
    def pins(self):
        """:obj:`list` of :class:`PadstackInstances <.PadstackInstance>`: \
        Padstack instances.

        This property is read-only.
        """
        ps = self.__stub.GetPins(self.msg).items
        return [PadstackInstance(p) for p in ps]

    def add_pins(self, pins):
        """Add a list of padstack instances to the group.

        Parameters
        ----------
        pins : list[:class:`.PadstackInstance`]
            List of padstick instances.
        """
        self.__stub.AddPins(messages.pin_group_pins_modify_message(self, pins))

    def remove_pins(self, pins):
        """Remove a list of padstack instances from the group.

        Parameters
        ----------
        pins : list[:class:`.PadstackInstance`]
            List of padstick instances.
        """
        self.__stub.RemovePins(messages.pin_group_pins_modify_message(self, pins))

    @property
    def pin_group_terminal(self):
        """:class:`.PinGroupTerminal`: Terminal this pin group is part of.

        This property is read-only.
        """
        return PinGroupTerminal(self.__stub.GetPinGroupTerminal(self.msg))



"""This module contains utility functions for API development work."""
from ansys.api.edb.v1.layout_obj_pb2 import LayoutObjTargetMessage

from ansys.edb.core.inner.factory import create_lyt_obj


def map_list(iterable_to_operate_on, operator=None):
    """Apply the given operator to each member of an iterable and return the modified list.

    Parameters
    ---------
    iterable_to_operate_on
    operator
    """
    return list(
        iterable_to_operate_on if operator is None else map(operator, iterable_to_operate_on)
    )


def query_lyt_object_collection(
    owner, obj_type, unary_rpc, unary_streaming_rpc, request_requires_type=True
):
    """For the provided request, retrieve a collection of objects using the unary_rpc or unary_streaming_rpc methods \
    depending on whether caching is enabled."""
    from ansys.edb.core.utility.io_manager import get_cache

    request = (
        LayoutObjTargetMessage(target=owner.msg, type=obj_type.value)
        if request_requires_type
        else owner.msg
    )

    items = []
    cache = get_cache()

    def add_msgs_to_items(edb_obj_collection_msg):
        nonlocal items
        for item in edb_obj_collection_msg.items:
            items.append(create_lyt_obj(item, obj_type))

    if cache is None:
        add_msgs_to_items(unary_rpc(request))
    else:
        for streamed_items in unary_streaming_rpc(request):
            add_msgs_to_items(streamed_items)
    return items


def stream_items_from_server(parser, stream, chunk_items_att_name):
    """Stream all items from the provided unary server stream and convert them to \
    the corresponding pyedb-core data type using the provided parser."""
    return [
        parser(chunk_entry)
        for chunk in stream
        for chunk_entry in getattr(chunk, chunk_items_att_name)
    ]


def client_stream_iterator(
    data_to_stream,
    chunk_type,
    chunk_entry_creator,
    chunk_entries_getter,
    max_size=32000,
    starting_chunk=None,
):
    """Create an iterator of client side items to be used when streaming items to the server."""
    chunk = chunk_type() if starting_chunk is None else starting_chunk
    for data_entry in data_to_stream:
        chunk_entry = chunk_entry_creator(data_entry)
        if chunk.ByteSize() + chunk_entry.ByteSize() > max_size:
            yield chunk
            chunk = chunk_type()
        chunk_entries_getter(chunk).append(chunk_entry)
    yield chunk



"""Layout object instance."""

from ansys.edb.core.inner import ObjBase, factory, parser, utils
from ansys.edb.core.inner.messages import bool_property_message, layer_ref_property_message
from ansys.edb.core.layer.layer import Layer
from ansys.edb.core.layout_instance import layout_instance_context
from ansys.edb.core.layout_instance.layout_obj_instance_2d_geometry import (
    LayoutObjInstance2DGeometry,
)
from ansys.edb.core.layout_instance.layout_obj_instance_3d_geometry import (
    LayoutObjInstance3DGeometry,
)
from ansys.edb.core.session import LayoutObjInstanceServiceStub, StubAccessor, StubType


def _parse_layout_obj_instance_geometry_message(lyt_obj_inst_geom_msg):
    if lyt_obj_inst_geom_msg.type == -1:
        raise TypeError("Encountered an unknown geometry type for the layout object instance.")
    geom_type = (
        LayoutObjInstance2DGeometry
        if lyt_obj_inst_geom_msg.type == 0
        else LayoutObjInstance3DGeometry
    )
    geom_msg = lyt_obj_inst_geom_msg.geometry
    return geom_type(geom_msg.geometry, geom_msg.owning_drawing, geom_msg.placement_layer)


class LayoutObjInstance(ObjBase):
    """Represents a layout object instance."""

    __stub: LayoutObjInstanceServiceStub = StubAccessor(StubType.layout_obj_instance)

    @property
    def layers(self):
        """:obj:`list` of :class:`.Layer`: All layer instances.

        This list contains the layer` instances that the layout object instance has geometry on.
        """
        return [Layer(msg).cast() for msg in self.__stub.GetLayers(self.msg).items]

    def get_geometries(self, layer):
        """Get the geometry that exists on a given layer.

        Parameters
        ----------
        layer : :class:`.Layer` or str
           Layer.

        Returns
        -------
        list[LayoutObjInstance2DGeometry or LayoutObjInstance3DGeometry]
        """
        geoms = self.__stub.GetGeometries(layer_ref_property_message(self, layer))
        return utils.map_list(geoms.geometries, _parse_layout_obj_instance_geometry_message)

    @property
    def context(self):
        r""":obj:`list`\[:obj:`str`\]: All strings representing the context of the layout object instance.

        This list of strings is a list of :class:`.CellInstance`
        names representing the hierarchy level this layout obj instance's :class:`context <.LayoutInstanceContext>`
        resides on. The first entry represents the top-level context and the last entry represents
        the context that the layout object instance exists in.

        This property is read-only.
        """
        return utils.map_list(self.__stub.GetContext(self.msg).strings)

    @property
    def layout_instance_context(self):
        """:class:`LayoutInstanceContext`: Context that the layout object instance exists in.

        This property is read-only.
        """
        return layout_instance_context.LayoutInstanceContext(
            self.__stub.GetLayoutInstanceContext(self.msg)
        )

    @property
    def layout_obj(self):
        """:term:`Connectable <Connectable>`: Definition layout object that the layout object \
        instance is an instance of.

        This property is read-only.
        """
        return factory.create_conn_obj(self.__stub.GetLayoutObj(self.msg))

    @parser.to_polygon_data
    def get_bbox(self, local=False):
        """Get the bounding box of the layout object instance.

        Parameters
        ----------
        local : bool
            Whether to return the bounding box in the local :class:`context <.LayoutInstanceContext>`.
            If ``False``, the bounding box is returned in the global context.

        Returns
        -------
        :class:`.PolygonData`
            Bounding box of the layout object instance.
        """
        return self.__stub.GetBBox(bool_property_message(self, local))



"""Port property."""

from ansys.api.edb.v1.port_property_pb2_grpc import PortPropertyServiceStub
import google.protobuf.empty_pb2 as empty_pb2

from ansys.edb.core.inner import ObjBase, messages
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.utility.value import Value


class PortProperty(ObjBase):
    """Represents a port property."""

    __stub: PortPropertyServiceStub = StubAccessor(StubType.port_property)

    @classmethod
    def create(cls):
        """
        Create a port property.

        Returns
        -------
        PortProperty
            Port property created.
        """
        return PortProperty(cls.__stub.Create(empty_pb2.Empty()))

    def clone(self):
        """
        Clone a port property.

        Returns
        -------
        PortProperty
            Port property cloned.
        """
        return PortProperty(self.__stub.Clone(messages.edb_obj_message(self)))

    @property
    def reference_height(self):
        """:class:`.Value`: Reference height of the port property.

        This property can be set with :term:`ValueLike`.
        """
        return Value(self.__stub.GetReferenceHeight(messages.edb_obj_message(self)))

    @reference_height.setter
    def reference_height(self, height):
        self.__stub.SetReferenceHeight(
            messages.value_property_message(self, messages.value_message(height))
        )

    @property
    def reference_size_auto(self):
        """:obj:`bool`: Flag indicating if the reference size is automatic."""
        return self.__stub.GetReferenceSizeAuto(messages.edb_obj_message(self)).value

    @reference_size_auto.setter
    def reference_size_auto(self, auto):
        self.__stub.SetReferenceSizeAuto(messages.bool_property_message(self, auto))

    def get_reference_size(self):
        r"""Get the X and Y reference sizes for the port property.

        Returns
        -------
        tuple[:class:`.Value`, :class:`.Value`]
        """
        value_pair_message = self.__stub.GetReferenceSize(messages.edb_obj_message(self))
        return Value(value_pair_message.val1), Value(value_pair_message.val2)

    def set_reference_size(self, ref_x, ref_y):
        """Set the X and Y reference sizes for the port property.

        Parameters
        ----------
        ref_x : :term:`ValueLike`
            X reference size.
        ref_y : :term:`ValueLike`
            Y reference size.
        """
        self.__stub.SetReferenceSize(messages.value_pair_property_message(self, ref_x, ref_y))



"""Transformations."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ansys.edb.core.typing import ValueLike
    from ansys.edb.core.geometry.point_data import PointData
    from ansys.edb.core.geometry.polygon_data import PolygonData

import ansys.api.edb.v1.transform_pb2 as pb
from ansys.api.edb.v1.transform_pb2_grpc import TransformServiceStub

from ansys.edb.core.inner import ObjBase, messages, parser
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.utility.value import Value


class Transform(ObjBase):
    """Represents a transformation."""

    __stub: TransformServiceStub = StubAccessor(StubType.transform)

    @classmethod
    def create(
        cls,
        scale: ValueLike,
        angle: ValueLike,
        mirror: bool,
        offset_x: ValueLike,
        offset_y: ValueLike,
    ) -> Transform:
        """Create a transform object.

        Parameters
        ----------
        scale : :term:`ValueLike`
            Scale parameter.
        angle : :term:`ValueLike`
            Rotation angle, specified counter-clockwise in radians.
        mirror : bool
            Mirror about Y-axis.
        offset_x : :term:`ValueLike`
            X offset.
        offset_y : :term:`ValueLike`
            Y offset.

        Returns
        -------
        .Transform
        """
        return Transform(
            cls.__stub.Create(
                pb.TransformMessage(
                    scale=messages.value_message(scale),
                    angle=messages.value_message(angle),
                    mirror=mirror,
                    offset_x=messages.value_message(offset_x),
                    offset_y=messages.value_message(offset_y),
                )
            )
        )

    @property
    def scale(self) -> Value:
        """:class:`.Value`: Scale property.

        This property can be set to :term:`ValueLike`.
        """
        return Value(self.__stub.GetScale(messages.edb_obj_message(self)))

    @scale.setter
    def scale(self, value: ValueLike):
        self.__stub.SetScale(messages.value_property_message(target=self, value=value))

    @property
    def rotation(self) -> Value:
        """:class:`.Value`: Rotation property.

        This property can be set to :term:`ValueLike`.
        """
        return Value(self.__stub.GetRotation(messages.edb_obj_message(self)))

    @rotation.setter
    def rotation(self, value: ValueLike):
        self.__stub.SetRotation(messages.value_property_message(target=self, value=value))

    @property
    def offset_x(self) -> Value:
        """:class:`.Value`: X offset property.

        This property can be set to :term:`ValueLike`.
        """
        return Value(self.__stub.GetOffsetX(messages.edb_obj_message(self)))

    @offset_x.setter
    def offset_x(self, value: ValueLike):
        self.__stub.SetOffsetX(messages.value_property_message(target=self, value=value))

    @property
    def offset_y(self) -> Value:
        """:class:`.Value`: Y offset property.

        This property can be set to :term:`ValueLike`.
        """
        return Value(self.__stub.GetOffsetY(messages.edb_obj_message(self)))

    @offset_y.setter
    def offset_y(self, value: ValueLike):
        self.__stub.SetOffsetY(messages.value_property_message(target=self, value=value))

    @property
    def mirror(self) -> Value:
        """:obj:`bool`: Mirror property. If ``True``, mirror about y axis."""
        return self.__stub.GetMirror(messages.edb_obj_message(self)).value

    @mirror.setter
    def mirror(self, value: ValueLike):
        self.__stub.SetMirror(messages.bool_property_message(self, value))

    @property
    def is_identity(self) -> bool:
        """:obj:`bool`: Flag indicating if the transformation is an identity transformation."""
        return self.__stub.IsIdentity(messages.edb_obj_message(self)).value

    def __add__(self, other_transform: Transform) -> Transform:
        """Add operator and concatenate two transformations.

        Parameters
        ----------
        other_transform : .Transform
            Second transformation

        Returns
        -------
        .Transform
            Transformation object created.
        """
        return Transform(
            self.__stub.TransformPlus(
                pb.TransformOperatorMessage(
                    target_1=messages.edb_obj_message(self.msg),
                    target_2=messages.edb_obj_message(other_transform.msg),
                )
            )
        )

    def transform_point(self, point: PointData):
        """Transform a point.

        Parameters
        ----------
        point : .PointData
            Point values [x, y] to transform.

        Returns
        -------
        .PointData
            Transformed point.
        """
        pnt_msg = self.__stub.TransformPoint(messages.point_property_message(self, point))
        return Value(pnt_msg.x), Value(pnt_msg.y)

    @parser.to_polygon_data
    def transform_polygon(self, polygon: PolygonData) -> PolygonData:
        """Transform a polygon.

        Parameters
        ----------
        polygon : .PolygonData
            Polygon to transform.

        Returns
        -------
        .PolygonData
            Transformed polygon.
        """
        return self.__stub.TransformPolygon(messages.polygon_data_property_message(self, polygon))



"""Differential pair."""

from ansys.edb.core.inner import messages
from ansys.edb.core.inner.layout_obj import LayoutObjType
from ansys.edb.core.net.net import Net
from ansys.edb.core.net.net_class import NetClass
from ansys.edb.core.session import DifferentialPairServiceStub, StubAccessor, StubType


class DifferentialPair(NetClass):
    """Represents a differential pair."""

    __stub: DifferentialPairServiceStub = StubAccessor(StubType.differential_pair)
    layout_obj_type = LayoutObjType.DIFFERENTIAL_PAIR

    @classmethod
    def create(cls, layout, name, pos_net=None, neg_net=None):
        """Create a differential pair.

        You must either provide both nets or omit both nets.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the differential pair in.
        name : str
            Name of the new differential pair.
        pos_net : Net or str, default: None
            Positive net or name of the positive net.
        neg_net : Net or str, default: None
            Negative net or the name of negative net.

        Returns
        -------
        DifferentialPair
            Differential pair created.
        """
        return cls(
            cls.__stub.Create(
                messages.differential_pair_creation_message(layout, name, pos_net, neg_net)
            )
        )

    @classmethod
    def find_by_name(cls, layout, name):
        """Find a differential pair by name in a given layout.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to search for the differential pair.
        name : str
            Name of the differential pair.

        Returns
        -------
        DifferentialPair
            Differential pair that was found. Check the :obj:`is_null
            <.DifferentialPair.is_null>`
            property of the returned differential pair to see if it exists.
        """
        return cls(cls.__stub.FindByName(messages.string_property_message(layout, name)))

    @property
    def differential_pair(self):
        r""":obj:`tuple`\(:class:`.Net`, :class:`.Net`\): Nets (positive, negative) in the differential pair."""
        msg = self.__stub.GetDifferentialPair(self.msg)
        return Net(msg.positive_net.id), Net(msg.negative_net.id)

    @differential_pair.setter
    def differential_pair(self, value):
        self.__stub.SetDifferentialPair(
            messages.differential_pair_net_refs_message(self, value[0], value[1])
        )

    @property
    def positive_net(self):
        """:class:`.Net`: Positive net in the differential pair."""
        return self.differential_pair[0]

    @positive_net.setter
    def positive_net(self, value):
        self.__stub.SetDifferentialPair(
            messages.differential_pair_net_refs_message(self, value, self.differential_pair[1])
        )

    @property
    def negative_net(self):
        """:class:`.Net`: Negative net in the differential pair."""
        return self.differential_pair[1]

    @negative_net.setter
    def negative_net(self, value):
        self.__stub.SetDifferentialPair(
            messages.differential_pair_net_refs_message(self, self.differential_pair[0], value)
        )

    @property
    def is_power_ground(self):
        """Flag indicating if the new is power/ground.

        This property is invalid for a differential pair.
        """
        return False

    def add_net(self, net):
        """Add a net.

        This method is invalid for a differential pair. Use
        :obj:`ansys.edb.core.net.differential_pair.DifferentialPair` = (pos_net, neg_net) instead.
        """
        raise TypeError("Net cannot be added to a differential pair.")

    def remove_net(self, net):
        """Remove a net.

        This method is invalid for a differential pair.
        """
        raise TypeError("Net cannot be removed from a differential pair.")

    @property
    def nets(self):
        """This property is invalid for a differential pair.

        Use :obj:`ansys.edb.core.net.differential_pair.DifferentialPair` instead.
        """
        return self.differential_pair



"""Adaptive solutions."""


class MatrixConvergenceDataEntry:
    """Class representing a matrix convergence data entry.

    Attributes
    ----------
    port_1_name : str
    port_2_name : str
    mag_limit : float
    phase_limit : float
    """

    def __init__(self, port_1_name, port_2_name, mag_limit, phase_limit):
        """Create a matrix convergence data entry."""
        self._port_1_name = port_1_name
        self._port_2_name = port_2_name
        self._mag_limit = mag_limit
        self._phase_limit = phase_limit

    @property
    def port_1_name(self):
        """:obj:`str`: Name of the first port."""
        return self._port_1_name

    @port_1_name.setter
    def port_1_name(self, port_1_name):
        self._port_1_name = port_1_name

    @property
    def port_2_name(self):
        """:obj:`str`: Name of the second port."""
        return self._port_2_name

    @port_2_name.setter
    def port_2_name(self, port_2_name):
        self._port_2_name = port_2_name

    @property
    def mag_limit(self):
        """:obj:`float`: Magnitude limit."""
        return self._mag_limit

    @mag_limit.setter
    def mag_limit(self, mag_limit):
        self._mag_limit = mag_limit

    @property
    def phase_limit(self):
        """:obj:`float`: Phase limit."""
        return self._phase_limit

    @phase_limit.setter
    def phase_limit(self, phase_limit):
        self._phase_limit = phase_limit


class MatrixConvergenceData:
    """Represents matrix convergence data for an adaptive frequency solution.

    Attributes
    ----------
    mag_min_threshold : float
    entry_list : list[:class:`.MatrixConvergenceDataEntry`]
    """

    def __init__(self, mag_min_threshold=0.01, entry_list=None):
        """Create a matrix convergence data entry."""
        self._all_constant = False
        self._all_diag_constant = False
        self._all_off_diag_constant = False
        self._mag_minThreshold = mag_min_threshold
        self._entry_list = entry_list if entry_list is not None else []

    @property
    def all_constant(self):
        """:obj:`bool`: Flag indicating if all matrix entries are the same.

        This property is read-only.
        """
        return self._all_constant

    @property
    def all_diag_constant(self):
        """:obj:`bool`: Flag indicating if all diagonal matrix entries are the same.

        This property is read-only.
        """
        return self._all_diag_constant

    @property
    def all_off_diag_constant(self):
        """:obj:`bool`: Flag indicating if all off-diagonal matrix entries are the same.

        This property is read-only.
        """
        return self._all_off_diag_constant

    @property
    def mag_min_threshold(self):
        """:obj:`float`: Minimum magnitude threshold.

        When the magnitude is less than the minimal threshold value, the phase is ignored.
        """
        return self._mag_minThreshold

    @mag_min_threshold.setter
    def mag_min_threshold(self, mag_min_threshold):
        self._mag_minThreshold = mag_min_threshold

    @property
    def entry_list(self):
        """:obj:`list` of :class:`.MatrixConvergenceDataEntry`: Matrix entries."""
        return self._entry_list

    @entry_list.setter
    def entry_list(self, entry_list):
        self._entry_list = entry_list
        self._all_constant = False
        self._all_diag_constant = False
        self._all_off_diag_constant = False

    def set_all_constant(self, mag_limit, phase_limit, port_names):
        """Set all entries in the matrix to have the given magnitude and phase values.

        Parameters
        ----------
        mag_limit : float
           Magnitude limit.
        phase_limit : float
            Phase limit.
        port_names : list[:obj:`str`]
            List of port names.
        """
        self._entry_list.clear()

        for port_name_1 in port_names:
            for port_name_2 in port_names:
                self._entry_list.append(
                    MatrixConvergenceDataEntry(port_name_1, port_name_2, mag_limit, phase_limit)
                )

        self._all_constant = True
        self._all_diag_constant = False
        self._all_off_diag_constant = False

    def set_all_diag_constant(self, mag_limit, phase_limit, port_names, clear_entries):
        """Set all diagonal entries in the matrix to have the given magnitude and phase values.

        Parameters
        ----------
        mag_limit : float
           Magnitude limit.
        phase_limit : float
            Phase limit.
        port_names : list[:obj:`str`]
            List of port names.
        clear_entries : bool
            Whether to clear entries.
        """
        if clear_entries:
            self._entry_list.clear()

        for port_name in port_names:
            self._entry_list.append(
                MatrixConvergenceDataEntry(port_name, port_name, mag_limit, phase_limit)
            )

        self._all_constant = False
        self._all_diag_constant = True
        if clear_entries:
            self._all_off_diag_constant = False

    def set_all_off_diag_constant(self, mag_limit, phase_limit, port_names, clear_entries):
        """Set all diagonal entries in the matrix to have the given magnitude and phase values.

        Parameters
        ----------
        mag_limit : float
           Magnitude limit.
        phase_limit : float
            Phase limit.
        port_names : list[:obj:`str`]
            List of port names.
        clear_entries : bool
            Whether to clear entries.
        """
        if clear_entries:
            self._entry_list.clear()

        for port_name_1 in port_names:
            for port_name_2 in port_names:
                if port_name_1 != port_name_2:
                    self._entry_list.append(
                        MatrixConvergenceDataEntry(port_name_1, port_name_2, mag_limit, phase_limit)
                    )

        self._all_constant = False
        if clear_entries:
            self._all_diag_constant = False
        self._all_off_diag_constant = True

    def add_entry(self, port_name_1, port_name_2, mag_limit, phase_limit):
        """Add a matrix entry.

        Parameters
        ----------
        port_name_1 : str
            Name of the first port.
        port_name_2 : str
            Name of the second port.
        mag_limit : float
            Magnitude limit.
        phase_limit : float
            Phase limit.
        """
        new_entry = MatrixConvergenceDataEntry(port_name_1, port_name_2, mag_limit, phase_limit)
        for idx, entry in enumerate(self._entry_list):
            if (
                entry.port_1_name == new_entry.port_1_name
                and entry.port_2_name == new_entry.port_2_name
            ):
                self._entry_list[idx] = new_entry
                return
        self._entry_list.append(new_entry)


class SingleFrequencyAdaptiveSolution:
    """Represents a single frequency adaptive solution.

    Attributes
    ----------
    adaptive_frequency : str
    max_delta : str
    max_passes : int
    mx_conv_data : MatrixConvergenceData
    use_mx_conv_data : bool
    """

    def __init__(
        self,
        adaptive_frequency="5GHz",
        max_delta="0.02",
        max_passes=10,
        mx_conv_data=None,
        use_mx_conv_data=False,
    ):
        """Create a single frequency adaptive solution."""
        self._adaptive_frequency = adaptive_frequency
        self._max_delta = max_delta
        self._max_passes = max_passes
        self._mx_conv_data = mx_conv_data if mx_conv_data is not None else MatrixConvergenceData()
        self._use_mx_conv_data = use_mx_conv_data

    @property
    def adaptive_frequency(self):
        """:obj:`str`: Adaptive frequency."""
        return self._adaptive_frequency

    @adaptive_frequency.setter
    def adaptive_frequency(self, adaptive_frequency):
        self._adaptive_frequency = adaptive_frequency

    @property
    def max_delta(self):
        """:obj:`str`: Maximum delta S between adaptive passes."""
        return self._max_delta

    @max_delta.setter
    def max_delta(self, max_delta):
        self._max_delta = max_delta

    @property
    def max_passes(self):
        """:obj:`int`: Maximum number of adaptive passes."""
        return self._max_passes

    @max_passes.setter
    def max_passes(self, max_passes):
        self._max_passes = max_passes

    @property
    def mx_conv_data(self):
        """:class:`.MatrixConvergenceData`: Matrix convergence data."""
        return self._mx_conv_data

    @mx_conv_data.setter
    def mx_conv_data(self, mx_conv_data):
        self._mx_conv_data = mx_conv_data

    @property
    def use_mx_conv_data(self):
        """:obj:`bool`: Flag indicating whether matrix convergence data is used."""
        return self._use_mx_conv_data

    @use_mx_conv_data.setter
    def use_mx_conv_data(self, use_mx_conv_data):
        self._use_mx_conv_data = use_mx_conv_data


class AdaptiveFrequency:
    """Represents an adaptive frequency.

    Attributes
    ----------
    adaptive_frequency : str
    max_delta : str
    output_variables : dict{ str : str }
    """

    def __init__(self, adaptive_frequency, max_delta="0.02", output_variables=None):
        """Create an adaptive frequency."""
        self._adaptive_frequency = adaptive_frequency
        self._max_delta = max_delta
        self._output_variables = output_variables if output_variables is not None else {}

    @property
    def adaptive_frequency(self):
        """:obj:`str`: Adaptive frequency."""
        return self._adaptive_frequency

    @adaptive_frequency.setter
    def adaptive_frequency(self, adaptive_frequency):
        self._adaptive_frequency = adaptive_frequency

    @property
    def max_delta(self):
        """:obj:`str`: Maximum delta S between adaptive passes."""
        return self._max_delta

    @max_delta.setter
    def max_delta(self, max_delta):
        self._max_delta = max_delta

    @property
    def output_variables(self):
        """:obj:`dict` { :obj:`str` : :obj:`str` }: Map of output variable names to maximum delta S."""
        return self._output_variables

    @output_variables.setter
    def output_variables(self, output_variables):
        self._output_variables = output_variables


class MultiFrequencyAdaptiveSolution:
    """Represents a multi-frequency adaptive solution.

    Attributes
    ----------
    max_passes : int
        Maximum number of adaptive passes.
    adaptive_frequencies : list[:class:`.AdaptiveFrequency`]
    """

    def __init__(self, max_passes=10, adaptive_frequencies=None):
        """Create a multi-frequency adaptive solution."""
        self._max_passes = max_passes
        if adaptive_frequencies is None or len(adaptive_frequencies) == 0:
            self._adaptive_frequencies = [
                AdaptiveFrequency(freq) for freq in ["2.5GHz", "5GHz", "10GHz"]
            ]
        else:
            self._adaptive_frequencies = adaptive_frequencies

    @property
    def adaptive_frequencies(self):
        """:obj:`list` of :class:`.AdaptiveFrequency`: Frequencies that adaptive solutions are calculated for."""
        return self._adaptive_frequencies

    @adaptive_frequencies.setter
    def adaptive_frequencies(self, adaptive_frequencies):
        self._adaptive_frequencies = adaptive_frequencies

    @property
    def max_passes(self):
        """:obj:`int`: Maximum number of adaptive passes."""
        return self._max_passes

    @max_passes.setter
    def max_passes(self, max_passes):
        self._max_passes = max_passes


class BroadbandAdaptiveSolution:
    """Represents a broadband adaptive solution.

    Attributes
    ----------
    low_frequency : str
    high_frequency : str
    max_num_passes : int
    max_delta : str
    """

    def __init__(
        self, low_frequency="5GHz", high_frequency="10GHz", max_num_passes=10, max_delta="0.02"
    ):
        """Create a broadband adaptive solution."""
        self._low_frequency = low_frequency
        self._high_frequency = high_frequency
        self._max_num_passes = max_num_passes
        self._max_delta = max_delta

    @property
    def low_frequency(self):
        """:obj:`str`: Lower-bound adaptive frequency."""
        return self._low_frequency

    @low_frequency.setter
    def low_frequency(self, low_frequency):
        self._low_frequency = low_frequency

    @property
    def high_frequency(self):
        """:obj:`str`: Higher-bound adaptive frequency."""
        return self._high_frequency

    @high_frequency.setter
    def high_frequency(self, high_frequency):
        self._high_frequency = high_frequency

    @property
    def max_num_passes(self):
        """:obj:`int`: Maximum number of adaptive passes."""
        return self._max_num_passes

    @max_num_passes.setter
    def max_num_passes(self, max_num_passes):
        self._max_num_passes = max_num_passes

    @property
    def max_delta(self):
        """:obj:`str`: Maximum delta S between adaptive passes."""
        return self._max_delta

    @max_delta.setter
    def max_delta(self, max_delta):
        self._max_delta = max_delta



"""3D transformformations."""
from __future__ import annotations

from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from ansys.edb.core.typing import Point3DLike
    from ansys.edb.core.geometry.point3d_data import Point3DData
    from ansys.edb.core.utility.transform import Transform

import ansys.api.edb.v1.transform_3d_pb2 as pb
from ansys.api.edb.v1.transform_3d_pb2_grpc import Transform3DServiceStub
from google.protobuf import empty_pb2

from ansys.edb.core.inner import ObjBase, messages
from ansys.edb.core.inner.parser import to_3_point3d_data, to_point3d_data
from ansys.edb.core.session import StubAccessor, StubType


class Transform3D(ObjBase):
    """Represents a 3D transformation.

    Parameters
    ----------
    anchor : :term:`Point3DLike`
    rot_axis_from : :term:`Point3DLike`
    rot_axis_to : :term:`Point3DLike`
    rot_angle : :term:`ValueLike`
        Rotation angle, specified counter-clockwise in radians, from the ``rot_axis_from`` parameter
        towards the ``rot_axis_to`` parameter.
    offset : :term:`Point3DLike`
    mirror : bool
        Mirror against the YZ plane.
    """

    __stub: Transform3DServiceStub = StubAccessor(StubType.transform3d)

    @classmethod
    def create_identity(cls) -> Transform3D:
        """Create an identity transformation 3D matrix.

        Returns
        -------
        .Transform3D
        """
        return Transform3D(cls.__stub.CreateIdentity(empty_pb2.Empty()))

    @classmethod
    def create_copy(cls, transform3d: Transform3D) -> Transform3D:
        """Create a 3D transformation by copying another 3D transformation.

        Returns
        -------
        .Transform3D
        """
        return Transform3D(cls.__stub.CreateCopy(messages.edb_obj_message(transform3d)))

    @classmethod
    def create_from_matrix(cls, matrix: List[List[float]]) -> Transform3D:
        """Create a 3D transformation from general matrix data.

        Parameters
        ----------
        matrix : :obj:`list` of :obj:`list` of :obj:`floats`
            Array (4x4) to copy from.

        Returns
        -------
        .Transform3D
        """
        trans = Transform3D(cls.__stub.CreateIdentity(empty_pb2.Empty()))
        trans.matrix = matrix
        return trans

    @classmethod
    def create_from_offset(cls, offset: Point3DLike) -> Transform3D:
        """Create a 3D transformation with an offset.

        Parameters
        ----------
        offset : :term:`Point3DLike`
            Vector offset.
        Returns
        -------
        .Transform3D
        """
        return Transform3D(cls.__stub.CreateOffset(messages.cpos_3d_message(offset)))

    @classmethod
    def create_from_center_scale(cls, center: Point3DLike, scale: float) -> Transform3D:
        """Create a 3D transformation for scaling about a point.

        Parameters
        ----------
        center : :term:`Point3DLike`
            Center of the transformation.
        scale : :obj:`float`
            Scale factor of the transformation.

        Returns
        -------
        .Transform3D
        """
        return Transform3D(
            cls.__stub.CreateCenterScale(messages.cpos_3d_double_message(center, scale))
        )

    @classmethod
    def create_from_angle(cls, zyx_decomposition: Point3DLike) -> Transform3D:
        """Create a 3D transformation from ZYX decomposition.

        Parameters
        ----------
        zyx_decomposition : :term:`Point3DLike`
            ZYX decomposition.

        Returns
        -------
        .Transform3D
        """
        return Transform3D(
            cls.__stub.CreateRotationFromAngle(messages.cpos_3d_message(zyx_decomposition))
        )

    @classmethod
    def create_from_axis(cls, x: Point3DLike, y: Point3DLike, z: Point3DLike) -> Transform3D:
        """Create a 3D transformation with a rotation matrix from three axes.

        Parameters
        ----------
        x : :term:`Point3DLike`
            X axis.
        y : :term:`Point3DLike`
            Y axis.
        z : :term:`Point3DLike`
            Z axis.

        Returns
        -------
        .Transform3D
        """
        return Transform3D(
            cls.__stub.CreateRotationFromAxis(messages.cpos_3d_triple_message(x, y, z))
        )

    @classmethod
    def create_from_axis_and_angle(cls, axis: Point3DLike, angle: float) -> Transform3D:
        """Create a 3D transformation with the given axis and angle.

        Parameters
        ----------
        axis : :term:`Point3DLike`
            Axis.
        angle : :obj:`float`
            Angle.

        Returns
        -------
        .Transform3D
        """
        return Transform3D(
            cls.__stub.CreateRotationFromAxisAndAngle(messages.cpos_3d_double_message(axis, angle))
        )

    @classmethod
    def create_from_one_axis_to_another(
        cls, from_axis: Point3DLike, to_axis: Point3DLike
    ) -> Transform3D:
        """Create a 3D transformformation with rotation from an axis to an axis.

        Parameters
        ----------
        from_axis : :term:`Point3DLike`
            From axis.
        to_axis : :term:`Point3DLike`
            To axis.

        Returns
        -------
        .Transform3D
        """
        return Transform3D(
            cls.__stub.CreateRotationFromToAxis(messages.cpos_3d_pair_message(from_axis, to_axis))
        )

    @classmethod
    def create_from_transform_2d(cls, transform: Transform, z_off: float) -> Transform3D:
        """Create a 3D transformation with transform data.

        Parameters
        ----------
        transform : .Transform
            Transform data.
        z_off : :obj:`float`
            Z offset.

        Returns
        -------
        .Transform3D
        """
        return Transform3D(
            cls.__stub.CreateTransform2D(messages.double_property_message(transform, z_off))
        )

    def transpose(self):
        """Transpose the 3D transformation."""
        self.__stub.Transpose(messages.edb_obj_message(self))

    def invert(self):
        """Invert the 3D transformation."""
        self.__stub.Invert(messages.edb_obj_message(self))

    def is_identity(self, eps: float, rotation: bool) -> bool:
        """Get identity of the 3D transformation.

        Parameters
        ----------
        rotation : bool
        eps : float

        Returns
        -------
        bool
        """
        return self.__stub.IsIdentity(
            pb.IsIdentityMessage(
                target=messages.edb_obj_message(self),
                eps=eps,
                rotation=rotation,
            )
        ).value

    def is_equal(self, other_transform: Transform3D, rotation: bool, eps: float) -> bool:
        """Equality check for two #D transformations.

        Parameters
        ----------
        other_transform
        rotation : bool
        eps : float

        Returns
        -------
        bool
            Result of equality check.
        """
        return self.__stub.IsEqual(
            pb.IsEqualMessage(
                target=messages.edb_obj_message(self),
                value=messages.edb_obj_message(other_transform),
                eps=eps,
                rotation=rotation,
            )
        ).value

    def __add__(self, other_transform: Transform3D) -> Transform3D:
        """Add operator and concatenate two 3D transformations.

        Parameters
        ----------
        other_transform : .Transform3D
            Second 3D transformation.

        Returns
        -------
        .Transform3D
            3D transformation object created.
        """
        return Transform3D(
            self.__stub.OperatorPlus(messages.pointer_property_message(self, other_transform))
        )

    @property
    @to_3_point3d_data
    def axis(self) -> Point3DData:
        """:obj:`list` of :term:`Point3DLike`: Axis."""
        return self.__stub.GetAxis(messages.edb_obj_message(self))

    @to_point3d_data
    def transform_point(self, point: Point3DLike) -> Point3DData:
        """Get the transform point of the 3D transformation.

        Parameters
        ----------
        point : :term:`Point3DLike`

        Returns
        -------
        .Point3DData
            Transform point.
        """
        return self.__stub.TransformPoint(messages.cpos_3d_property_message(self, point))

    @property
    @to_point3d_data
    def z_y_x_rotation(self) -> Point3DData:
        """:class:`.Point3DData`: ZYX rotation.

        This property is read-only.
        """
        return self.__stub.GetZYXRotation(messages.edb_obj_message(self))

    @property
    @to_point3d_data
    def scaling(self) -> Point3DData:
        """:class:`.Point3DData`: Scaling.

        This property is read-only.
        """
        return self.__stub.GetScaling(messages.edb_obj_message(self))

    @property
    @to_point3d_data
    def shift(self) -> Point3DData:
        """:class:`.Point3DData`: Shift.

        This property is read-only.
        """
        return self.__stub.GetShift(messages.edb_obj_message(self))

    @property
    def matrix(self) -> List[List[float]]:
        """:obj:`list` of :obj:`list` of :obj:`float` : Transformation matrix as a 2D 4x4 array."""
        msg = self.__stub.GetMatrix(messages.edb_obj_message(self))
        matrix = [[float(_) for _ in msg.doubles[(i - 1) * 4 : i * 4]] for i in range(1, 5)]
        return matrix

    @matrix.setter
    def matrix(self, value: List[List[float]]):
        if (
            len(value) == 4
            and len(value[0]) == 4
            and len(value[1]) == 4
            and len(value[2]) == 4
            and len(value[3]) == 4
        ):
            unrolled_matrix = [float(j) for submatrix in value for j in submatrix]
            self.__stub.SetMatrix(messages.doubles_property_message(self, unrolled_matrix))



"""Circle."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ansys.edb.core.layout.layout import Layout
    from ansys.edb.core.geometry.polygon_data import PolygonData
    from ansys.edb.core.typing import NetLike, LayerLike, ValueLike

from ansys.api.edb.v1 import circle_pb2, circle_pb2_grpc

from ansys.edb.core.inner import messages, parser
from ansys.edb.core.primitive.primitive import Primitive
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.utility.value import Value


class Circle(Primitive):
    """Represents a circle object."""

    __stub: circle_pb2_grpc.CircleServiceStub = StubAccessor(StubType.circle)

    @classmethod
    def create(
        cls,
        layout: Layout,
        layer: LayerLike,
        net: NetLike | None,
        center_x: ValueLike,
        center_y: ValueLike,
        radius: ValueLike,
    ) -> Circle:
        """Create a circle.

        Parameters
        ----------
        layout : .Layout
            Layout to create this circle in.
        layer : :term:`LayerLike`
            Layer to place the circle on.
        net : :term:`NetLike` or None
            Net of the circle.
        center_x : :term:`ValueLike`
            X value of the center point.
        center_y : :term:`ValueLike`
            Y value of the center point.
        radius : :term:`ValueLike`
            Radius value of the circle.

        Returns
        -------
        Circle
            Circle created.
        """
        return Circle(
            cls.__stub.Create(
                circle_pb2.CircleCreationMessage(
                    layout=layout.msg,
                    layer=messages.layer_ref_message(layer),
                    net=messages.net_ref_message(net),
                    center_x=messages.value_message(center_x),
                    center_y=messages.value_message(center_y),
                    radius=messages.value_message(radius),
                )
            )
        )

    @classmethod
    @parser.to_polygon_data
    def render(
        cls, center_x: ValueLike, center_y: ValueLike, radius: ValueLike, is_hole: bool
    ) -> PolygonData:
        """Render a circle.

        Parameters
        ----------
        center_x : :term:`ValueLike`
            X value of the center point.
        center_y : :term:`ValueLike`
            Y value of the center point.
        radius : :term:`ValueLike`
            Radius value of the circle.
        is_hole : bool
            Whether the circle object is a hole.

        Returns
        -------
        .PolygonData
            Circle created.
        """
        return cls.__stub.Render(
            circle_pb2.CircleRenderMessage(
                center_x=messages.value_message(center_x),
                center_y=messages.value_message(center_y),
                radius=messages.value_message(radius),
                is_hole=is_hole,
            )
        )

    def get_parameters(self) -> tuple[Value, Value, Value]:
        """Get parameters of the circle.

        Returns
        -------
        tuple of (.Value, .Value, .Value)

            Returns a tuple in this format:

            **(center_x, center_y, radius)**

            **center_x** : X value of center point.

            **center_y** : Y value of center point.

            **radius** : Radius value of the circle.
        """
        circle_param_msg = self.__stub.GetParameters(self.msg)
        return (
            Value(circle_param_msg.center_x),
            Value(circle_param_msg.center_y),
            Value(circle_param_msg.radius),
        )

    def set_parameters(self, center_x: ValueLike, center_y: ValueLike, radius: ValueLike):
        """Set parameters of the circle.

        Parameters
        ----------
        center_x : :term:`ValueLike`
            X value of the center point.
        center_y : :term:`ValueLike`
            Y value of the center point.
        radius : :term:`ValueLike`
            Radius value of the circle.
        """
        self.__stub.SetParameters(
            circle_pb2.SetCircleParametersMessage(
                target=self.msg,
                parameters=circle_pb2.CircleParametersMessage(
                    center_x=messages.value_message(center_x),
                    center_y=messages.value_message(center_y),
                    radius=messages.value_message(radius),
                ),
            )
        )

    @property
    def polygon_data(self) -> PolygonData:
        """:class:`.PolygonData`: \
        Polygon data object of the circle.

        This property is read-only.
        """
        return Circle.render(*self.get_parameters(), self.is_void)

    def can_be_zone_primitive(self) -> bool:
        """:obj:`bool`: Flag indicating if a circle can be a zone."""
        return True



"""RaptorX simulation setup."""

from ansys.edb.core.simulation_setup.raptor_x_simulation_settings import RaptorXSimulationSettings
from ansys.edb.core.simulation_setup.simulation_setup import SimulationSetup, SimulationSetupType


class RaptorXSimulationSetup(SimulationSetup):
    """Represents RaptorX simulation setup data."""

    @classmethod
    def create(cls, cell, name):
        """Create a RaptorX simulation setup.

        Parameters
        ----------
        cell : :class:`.Cell`
            Cell to create the simulation setup in.
        name : str
            Name of the simulation setup.

        Returns
        -------
        RaptorXSimulationSetup
            RaptorX simulation setup created.
        """
        return super()._create(cell, name, SimulationSetupType.RAPTOR_X)

    @property
    def settings(self):
        """:class:`.RaptorXSimulationSettings`: Simulation settings of the RaptorX simulation setup."""
        return RaptorXSimulationSettings(self)



"""Arc data."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ansys.edb.core.typing import PointLike


from enum import Enum
import math

from ansys.api.edb.v1 import arc_data_pb2_grpc

from ansys.edb.core import session
from ansys.edb.core.geometry.point_data import PointData
from ansys.edb.core.inner import messages, parser
from ansys.edb.core.utility import conversions


class RotationDirection(Enum):
    """Represents arc directions."""

    CW = "cw"
    CCW = "ccw"
    CO_LINEAR = "colinear"


class ArcData:
    """Represents arc data."""

    __stub: arc_data_pb2_grpc.ArcDataServiceStub = session.StubAccessor(session.StubType.arc_data)

    def __init__(self, start: PointLike, end: PointLike, **kwargs):
        """Create an arc.

        Parameters
        ----------
        start : :term:`Point2DLike`
        end : :term:`Point2DLike`
        height: float, int, optional
        direction : Literal["cw", "ccw", "colinear"], optional
        """
        self._start = conversions.to_point(start)
        self._end = conversions.to_point(end)
        self._height = None
        self._height_options = kwargs

        if "height" in kwargs or len(kwargs) == 0:
            self._height = kwargs.get("height", 0.0)
        if "direction" in kwargs:
            self._height_options["direction"] = RotationDirection(kwargs["direction"])

    def __str__(self) -> str:
        """Generate a readable string for the arc.

        Returns
        -------
        str
        """
        if self.height == 0:
            return f"{self.start} {self.end}"
        else:
            arc = PointData(self.height)
            return f"{self.start} {arc} {self.end}"

    @property
    def start(self) -> PointData:
        """
        :class:`.PointData`: Start point of the arc.

        This property is read-only.
        """
        return self._start

    @property
    def end(self) -> PointData:
        """
        :class:`.PointData`: End point of the arc.

        This property is read-only.
        """
        return self._end

    @property
    def height(self) -> float:
        """
        :obj:`float`: Height of the arc.

        This property is read-only.
        """
        if self._height is None:
            self._height = self.__stub.GetHeight(messages.arc_message(self)).value

        return self._height

    def is_point(self, tolerance: float = 0.0) -> bool:
        """Determine if the arc is a point.

        An arc is a point when its start and end points are the same.

        Parameters
        ----------
        tolerance : float, default: 0.0
           Tolearance.

        Returns
        -------
        bool
            ``True`` when the arc is a point, ``False`` otherwise.
        """
        return self.is_segment(tolerance) and self.start.equals(self.end, tolerance)

    def is_segment(self, tolerance: float = 0.0) -> bool:
        """Determine if the arc is a straight line segment.

        Parameters
        ----------
        tolerance : float, default: 0.0
            Tolearance.

        Returns
        -------
        bool
            ``True`` when the arc is a straight line segment, ``False`` otherwise.
        """
        return math.fabs(self.height) <= tolerance

    @property
    @parser.to_point_data
    def center(self) -> PointData:
        """
        :class:`.PointData`: Center point of the arc.

        This property is read-only.
        """
        return self.__stub.GetCenter(messages.arc_message(self))

    @property
    @parser.to_point_data
    def midpoint(self) -> PointData:
        """
        :class:`.PointData`: Midpoint of the arc.

        This property is read-only.
        """
        return self.__stub.GetMidpoint(messages.arc_message(self))

    @property
    def radius(self) -> float:
        """
        :obj:`float`: Radius of the arc.

        This property is read-only.
        """
        return self.__stub.GetRadius(messages.arc_message(self)).value

    @property
    @parser.to_polygon_data
    def bbox(self) -> PolygonData:
        """
        :class:`.PolygonData`: Rectangular bounding box of the arc.

        This property is read-only.
        """
        return self.__stub.GetBoundingBox(messages.arc_message(self))

    def is_big(self) -> bool:
        """Determine if the arc is big.

        Returns
        -------
        bool
            ``True`` when the arc is big, ``False`` otherwise.
        """
        dist = self.start.distance(self.end)
        return 2 * math.fabs(self.height) > dist

    def is_left(self) -> bool:
        """Determine if the arc rotates clockwise.

        This method is the same as the :obj:`is_cw` method.

        Returns
        -------
        bool
            ``True`` when the arc rotates clockwise, ``False`` otherwise.
        """
        return self.is_cw()

    def is_cw(self) -> bool:
        """Determine if the arc rotates clockwise.

        This method is the same as the ``is_left`` method.

        Returns
        -------
        bool
            ``True`` when the arc rotates clockwise, ``False`` otherwise.
        """
        return self.height > 0.0

    def is_ccw(self) -> bool:
        """Determine if the arc rotates counter-clockwise.

        Returns
        -------
        bool
            ``True`` when the arc rotates counter-clockwise, ``False`` otherwise.
        """
        return self.height < 0.0

    @property
    def direction(self) -> str:
        """
        :obj:`str` : Rotational direction of the arc which can be "cw" or "ccw" or "colinear".

        This property is read-only.
        """
        if self.is_cw():
            return "cw"
        elif self.is_ccw():
            return "ccw"
        else:
            return "colinear"

    def angle(self, arc: ArcData = None) -> float:
        """Get the angle between this arc and another arc if provided or the angle of this arc.

        Parameters
        ----------
        arc : ArcData, default: None
           Other arc.

        Returns
        -------
        float
            Angle in radians.
        """
        if arc is None:
            return self.__stub.GetAngle(messages.arc_message(self)).value

        if self.is_segment() and arc.is_segment():
            vec1 = self.end - self.start
            vec2 = arc.end - arc.start
        else:
            point1, point2 = self.closest_points(arc)
            vec1, vec2 = self.tangent_at(point1), self.tangent_at(point2)

        return vec1.angle(vec2)

    @property
    def length(self) -> str:
        """
        :obj:`str`: Circumference length of the arc.

        This property is read-only.
        """
        if self.is_segment():
            return self.start.distance(self.end)
        else:
            return math.fabs(self.angle() * self.radius)

    @property
    def points(self) -> list[PointData]:
        """
        :obj:`list` of :class:`.PointData`: Geometric points representing the arc.

        This property is read-only.
        """
        return [self._start, PointData(self.height), self._end]

    def tangent_at(self, point: PointLike) -> PointData:
        """Get the tangent vector of the arc at a given point.

        Parameters
        ----------
        point : :term:`Point2DLike`
            Point.

        Returns
        -------
        .PointData
        """
        if self.is_segment():
            return self.end - self.start

        point = conversions.to_point(point)
        vec = point - self.center

        if self.is_ccw():
            return PointData(-vec.y, vec.x)
        else:
            return PointData(vec.y, -vec.x)

    @parser.to_box
    def closest_points(self, other: ArcData) -> tuple[PointData, PointData]:
        """Get the closest points from this arc to another arc, and vice versa.

        Parameters
        ----------
        other : .ArcData
            Other arc.

        Returns
        -------
        tuple of (.PointData, .PointData)
        """
        return self.__stub.ClosestPoints(messages.arc_data_two_points(self, other))



"""Q3D simulation settings."""
from __future__ import annotations

from enum import Enum

import ansys.api.edb.v1.hfss_simulation_settings_pb2 as hfss_pb2
import ansys.api.edb.v1.q3d_simulation_settings_pb2 as pb

from ansys.edb.core.inner import messages
from ansys.edb.core.session import (
    Q3DAdvancedMeshingSettingsServiceStub,
    Q3DAdvancedSettingsServiceStub,
    Q3DCGSettingsServiceStub,
    Q3DDCRLSettingsServiceStub,
    Q3DGeneralSettingsServiceStub,
    Q3DSettingsServiceStub,
    StubAccessor,
    StubType,
)
from ansys.edb.core.simulation_setup.hfss_simulation_settings import SolverType
from ansys.edb.core.simulation_setup.simulation_settings import (
    AdvancedMeshingSettings,
    AdvancedSettings,
    SimulationSettings,
    SimulationSettingsBase,
)


class Q3DSolutionOrder(Enum):
    """Provides an enum representing Q3D solution order."""

    NORMAL = pb.Q3DSolutionOrder.NORMAL
    HIGH = pb.Q3DSolutionOrder.HIGH
    HIGHER = pb.Q3DSolutionOrder.HIGHER
    HIGHEST = pb.Q3DSolutionOrder.HIGHEST
    NUM_SOLUTION_ORDER = pb.Q3DSolutionOrder.NUM_SOLUTION_ORDER


class Q3DSettingsType(Enum):
    """Provides an enum representing Q3D settings type."""

    DC = pb.Q3DSettingsType.DC
    AC = pb.Q3DSettingsType.AC
    CG = pb.Q3DSettingsType.CG


class Q3DSimulationSettings(SimulationSettings):
    """Represents Q3D simulation settings."""

    @property
    def general(self) -> Q3DGeneralSettings:
        """:class:`.Q3DGeneralSettings`: General settings for Q3D simulations."""
        return Q3DGeneralSettings(self._sim_setup)

    @property
    def dcrl(self) -> Q3DDCRLSettings:
        """:class:`.Q3DDCRLSettings`: DCRL settings for Q3D simulations."""
        return Q3DDCRLSettings(self._sim_setup)

    @property
    def acrl(self) -> Q3DACRLSettings:
        """:class:`.Q3DACRLSettings`: ACRL settings for Q3D simulations."""
        return Q3DACRLSettings(self._sim_setup)

    @property
    def cg(self) -> Q3DCGSettings:
        """:class:`.Q3DCGSettings`: CG settings for Q3D simulations."""
        return Q3DCGSettings(self._sim_setup)

    @property
    def advanced(self) -> Q3DAdvancedSettings:
        """:class:`.Q3DAdvancedSettings`: Advanced settings for Q3D simulations."""
        return Q3DAdvancedSettings(self._sim_setup)

    @property
    def advanced_meshing(self) -> Q3DAdvancedMeshingSettings:
        """:class:`.Q3DAdvancedMeshingSettings`: Advanced meshing settings for Q3D simulations."""
        return Q3DAdvancedMeshingSettings(self._sim_setup)


class Q3DGeneralSettings(SimulationSettingsBase):
    """Represents general settings for Q3D simulations."""

    __stub: Q3DGeneralSettingsServiceStub = StubAccessor(StubType.q3d_general_sim_settings)

    @property
    def solution_frequency(self) -> str:
        """:`obj`:str: Simulation frequency for Q3D."""
        return self.__stub.GetSolutionFrequency(self.msg).value

    @solution_frequency.setter
    def solution_frequency(self, solution_frequency: str):
        self.__stub.SetSolutionFrequency(messages.string_property_message(self, solution_frequency))

    @property
    def do_dc(self) -> bool:
        """:`obj`:bool: Whether to solve and generate mesh for for DC resistance and inductance."""
        return self.__stub.GetDoDC(self.msg).value

    @do_dc.setter
    def do_dc(self, do_dc: bool):
        self.__stub.SetDoDC(messages.bool_property_message(self, do_dc))

    @property
    def do_dc_res_only(self) -> bool:
        """:`obj`:bool: Whether to perform DC analysis with resistive only."""
        return self.__stub.GetDoDCResOnly(self.msg).value

    @do_dc_res_only.setter
    def do_dc_res_only(self, do_dc_res_only: bool):
        self.__stub.SetDoDCResOnly(messages.bool_property_message(self, do_dc_res_only))

    @property
    def do_cg(self) -> bool:
        """:`obj`:bool: Whether to solve and generate mesh for DC capacitance and conductance."""
        return self.__stub.GetDoCG(self.msg).value

    @do_cg.setter
    def do_cg(self, do_cg: bool):
        self.__stub.SetDoCG(messages.bool_property_message(self, do_cg))

    @property
    def do_ac(self) -> bool:
        """:`obj`:bool: Whether to p to solve and generate mesh for AC resistance and inductance."""
        return self.__stub.GetDoAC(self.msg).value

    @do_ac.setter
    def do_ac(self, do_ac: bool):
        self.__stub.SetDoAC(messages.bool_property_message(self, do_ac))

    @property
    def save_fields(self) -> bool:
        """:`obj`:bool: Whether to save the current settings as the new default settings."""
        return self.__stub.GetSaveFields(self.msg).value

    @save_fields.setter
    def save_fields(self, save_fields: bool):
        self.__stub.SetSaveFields(messages.bool_property_message(self, save_fields))


class Q3DSettings(SimulationSettingsBase):
    """Represents settings for Q3D simulations."""

    __stub: Q3DSettingsServiceStub = StubAccessor(StubType.q3d_sim_settings)

    def __init__(self, sim_setup, sim_type):
        """Create a Q3DSettings object."""
        super().__init__(sim_setup)
        self._sim_type = sim_type

    @property
    def max_passes(self) -> int:
        """:`obj`:int: Maximum number of mesh refinement cycles to perform."""
        return self.__stub.GetMaxPasses(self._q3d_settings_message()).value

    @max_passes.setter
    def max_passes(self, max_passes: int):
        self.__stub.SetMaxPasses(self._q3d_settings_uint64_property_message(max_passes))

    @property
    def min_passes(self) -> int:
        """:`obj`:int: Minimum number of mesh refinement cycles to perform."""
        return self.__stub.GetMinPasses(self._q3d_settings_message()).value

    @min_passes.setter
    def min_passes(self, min_passes: int):
        self.__stub.SetMinPasses(self._q3d_settings_uint64_property_message(min_passes))

    @property
    def min_converged_passes(self) -> int:
        """:`obj`:int: Minimum number of passes that must meet convergence criteria before the simulation stops."""
        return self.__stub.GetMinConvergedPasses(self._q3d_settings_message()).value

    @min_converged_passes.setter
    def min_converged_passes(self, min_converged_passes: int):
        self.__stub.SetMinConvergedPasses(
            self._q3d_settings_uint64_property_message(min_converged_passes)
        )

    @property
    def percent_error(self) -> float:
        """:obj:`float`: Percent error during conduction adaptive passes."""
        return self.__stub.GetPercentError(self._q3d_settings_message()).value

    @percent_error.setter
    def percent_error(self, percent_error: float):
        self.__stub.SetPercentError(self._q3d_settings_double_property_message(percent_error))

    @property
    def max_refine_per_pass(self) -> float:
        """:`obj`:float: How many tetrahedra are added at each iteration of the adaptive refinement process."""
        return self.__stub.GetMaxRefinePerPass(self._q3d_settings_message()).value

    @max_refine_per_pass.setter
    def max_refine_per_pass(self, max_refine_per_pass: float):
        self.__stub.SetMaxRefinePerPass(
            self._q3d_settings_double_property_message(max_refine_per_pass)
        )

    def _q3d_settings_message(self):
        return pb.Q3DSettingsMessage(target=self.msg, q3d_settings_type=self._sim_type.value)

    def _q3d_settings_uint64_property_message(self, value):
        return pb.Q3DSettingsUInt64PropertyMessage(
            q3d_settings=self._q3d_settings_message(), value=value
        )

    def _q3d_settings_double_property_message(self, value):
        return pb.Q3DSettingsDoublePropertyMessage(
            q3d_settings=self._q3d_settings_message(), value=value
        )


class Q3DDCRLSettings(Q3DSettings):
    """Represents DCRL settings for Q3D simulations."""

    __stub: Q3DDCRLSettingsServiceStub = StubAccessor(StubType.q3d_dcrl_sim_settings)

    def __init__(self, sim_setup):
        """Create a Q3DDCRLSettings object."""
        super().__init__(sim_setup, Q3DSettingsType.DC)

    @property
    def solution_order(self) -> Q3DSolutionOrder:
        """:class:`.Q3DSolutionOrder`: Inductance adaptive solution accuracy. \
            The solver will give accurate results at the Normal setting for most applications."""
        return Q3DSolutionOrder(self.__stub.GetSolutionOrder(self.msg).q3d_solution_order)

    @solution_order.setter
    def solution_order(self, q3d_solution_order: Q3DSolutionOrder):
        self.__stub.SetSolutionOrder(
            pb.Q3DSolutionOrderPropertyMessage(
                target=self.msg, q3d_solution_order=q3d_solution_order.value
            )
        )


class Q3DACRLSettings(Q3DSettings):
    """Represents ACRL settings for Q3D simulations."""

    def __init__(self, sim_setup):
        """Create a Q3DACRLSettings object."""
        super().__init__(sim_setup, Q3DSettingsType.AC)


class Q3DCGSettings(Q3DSettings):
    """Represents CG settings for Q3D simulations."""

    __stub: Q3DCGSettingsServiceStub = StubAccessor(StubType.q3d_cg_sim_settings)

    def __init__(self, sim_setup):
        """Create a Q3DCGSettings object."""
        super().__init__(sim_setup, Q3DSettingsType.CG)

    @property
    def auto_incr_sol_order(self) -> bool:
        """:`obj`:bool: Whether to automatically increase accuracy."""
        return self.__stub.GetAutoIncrSolOrder(self.msg).value

    @auto_incr_sol_order.setter
    def auto_incr_sol_order(self, auto_incr_sol_order: bool):
        self.__stub.SetAutoIncrSolOrder(messages.bool_property_message(self, auto_incr_sol_order))

    @property
    def solution_order(self) -> Q3DSolutionOrder:
        """:class:`.Q3DSolutionOrder`: Adaptive solution accuracy. \
            The solver will give accurate results at the Normal setting for most applications."""
        return Q3DSolutionOrder(self.__stub.GetSolutionOrder(self.msg).q3d_solution_order)

    @solution_order.setter
    def solution_order(self, q3d_solution_order: Q3DSolutionOrder):
        self.__stub.SetSolutionOrder(
            pb.Q3DSolutionOrderPropertyMessage(
                target=self.msg, q3d_solution_order=q3d_solution_order.value
            )
        )

    @property
    def solver_type(self):
        """:class:`.SolverType`: Q3D solver type."""
        return SolverType(self.__stub.GetSolverType(self.msg).solver_type)

    @solver_type.setter
    def solver_type(self, solver_type):
        self.__stub.SetSolverType(
            hfss_pb2.SolverTypePropertyMessage(target=self.msg, solver_type=solver_type.value)
        )

    @property
    def compression_tol(self) -> float:
        """:`obj`:float: Compression tolerance value."""
        return self.__stub.GetCompressionTol(self.msg).value

    @compression_tol.setter
    def compression_tol(self, compression_tol: float):
        self.__stub.SetCompressionTol(messages.double_property_message(self, compression_tol))


class Q3DAdvancedSettings(AdvancedSettings, Q3DSettings):
    """Represents advanced settings for Q3D simulations."""

    __stub: Q3DAdvancedSettingsServiceStub = StubAccessor(StubType.q3d_advanced_sim_settings)

    @property
    def ic_mode_auto_resolution(self) -> bool:
        """:obj:`bool`: Flag indicating if model resolution is automatically calculated for IC designs."""
        return self.__stub.GetICModeAutoResolution(self.msg).value

    @ic_mode_auto_resolution.setter
    def ic_mode_auto_resolution(self, ic_mode_auto_resolution: bool):
        self.__stub.SetICModeAutoResolution(
            messages.bool_property_message(self, ic_mode_auto_resolution)
        )

    @property
    def ic_mode_length(self) -> str:
        """:obj:`str`: Model resolution to use when manually setting the model resolution of IC designs."""
        return self.__stub.GetICModeLength(self.msg).value

    @ic_mode_length.setter
    def ic_mode_length(self, ic_mode_length: str):
        self.__stub.SetICModeLength(messages.string_property_message(self, ic_mode_length))


class Q3DAdvancedMeshingSettings(AdvancedMeshingSettings):
    """Represents advanced meshing settings for Q3D simulations."""

    __stub: Q3DAdvancedMeshingSettingsServiceStub = StubAccessor(
        StubType.q3d_advanced_meshing_sim_settings
    )

    @property
    def layer_alignment(self) -> str:
        """:obj:`str`: Snapping tolerance for hierarchical layer alignment.

        Unitless values represent a fraction of the total stackup height.
        """
        return self.__stub.GetLayerAlignment(self.msg).value

    @layer_alignment.setter
    def layer_alignment(self, layer_alignment: str):
        self.__stub.SetLayerAlignment(messages.bool_property_message(self, layer_alignment))



"""Layout object instance 3D geometry."""

from ansys.edb.core.geometry.triangle3d_data import Triangle3DData
from ansys.edb.core.inner import utils
from ansys.edb.core.inner.parser import to_point3d_data
from ansys.edb.core.layout_instance.layout_obj_instance_geometry import LayoutObjInstanceGeometry
from ansys.edb.core.session import LayoutObjInstance3DGeometryServiceStub, StubAccessor, StubType


class LayoutObjInstance3DGeometry(LayoutObjInstanceGeometry):
    """Represents a layout object instance 3D geometry."""

    __stub: LayoutObjInstance3DGeometryServiceStub = StubAccessor(
        StubType.layout_obj_instance_3d_geometry
    )

    @property
    def tesselation_data(self):
        """:obj:`list` of :class:`.Triangle3DData`: All triangle 3D data instances.

        This list contains triangle 3D data instances that correspond to the
        underlying tessellation data of the geometry.
        """
        tesselation_data = self.__stub.GetTesselationData(self.msg)

        def to_3d_triangle(triangle_msg):
            return Triangle3DData(
                to_point3d_data(triangle_msg.point_1),
                to_point3d_data(triangle_msg.point_2),
                to_point3d_data(triangle_msg.point_3),
            )

        return utils.map_list(tesselation_data.tesselation_data, to_3d_triangle)



"""Session manager for gRPC."""
from __future__ import annotations

from collections import defaultdict
from contextlib import contextmanager
from enum import Enum
import errno
import json
import os
from shutil import which
import socket
from struct import pack, unpack
import subprocess
from sys import modules
from typing import List

from ansys.api.edb.v1.arc_data_pb2_grpc import ArcDataServiceStub
from ansys.api.edb.v1.board_bend_def_pb2_grpc import BoardBendDefServiceStub
from ansys.api.edb.v1.bondwire_def_pb2_grpc import (
    ApdBondwireDefServiceStub,
    BondwireDefServiceStub,
    Jedec4BondwireDefServiceStub,
    Jedec5BondwireDefServiceStub,
)
from ansys.api.edb.v1.bondwire_pb2_grpc import BondwireServiceStub
from ansys.api.edb.v1.bundle_term_pb2_grpc import BundleTerminalServiceStub
from ansys.api.edb.v1.cell_instance_pb2_grpc import CellInstanceServiceStub
from ansys.api.edb.v1.cell_pb2_grpc import CellServiceStub
from ansys.api.edb.v1.circle_pb2_grpc import CircleServiceStub
from ansys.api.edb.v1.component_def_pb2_grpc import ComponentDefServiceStub
from ansys.api.edb.v1.component_group_pb2_grpc import ComponentGroupServiceStub
from ansys.api.edb.v1.component_model_pb2_grpc import (
    ComponentModelServiceStub,
    DynamicLinkComponentModelServiceStub,
    NPortComponentModelServiceStub,
)
from ansys.api.edb.v1.component_pin_pb2_grpc import ComponentPinServiceStub
from ansys.api.edb.v1.component_property_pb2_grpc import ComponentPropertyServiceStub
from ansys.api.edb.v1.connectable_pb2_grpc import ConnectableServiceStub
from ansys.api.edb.v1.database_pb2_grpc import DatabaseServiceStub
from ansys.api.edb.v1.dataset_def_pb2_grpc import DatasetDefServiceStub
from ansys.api.edb.v1.debye_model_pb2_grpc import DebyeModelServiceStub
from ansys.api.edb.v1.die_property_pb2_grpc import DiePropertyServiceStub
from ansys.api.edb.v1.dielectric_material_model_pb2_grpc import DielectricMaterialModelServiceStub
from ansys.api.edb.v1.differential_pair_pb2_grpc import DifferentialPairServiceStub
from ansys.api.edb.v1.djordjecvic_sarkar_model_pb2_grpc import DjordjecvicSarkarModelServiceStub
from ansys.api.edb.v1.edb_error_manager_pb2_grpc import EDBErrorManagerServiceStub
from ansys.api.edb.v1.edge_term_pb2_grpc import EdgeServiceStub, EdgeTerminalServiceStub
from ansys.api.edb.v1.extended_net_pb2_grpc import ExtendedNetServiceStub
from ansys.api.edb.v1.group_pb2_grpc import GroupServiceStub
from ansys.api.edb.v1.hfss_pi_simulation_settings_pb2_grpc import (
    HFSSPIGeneralSettingsServiceStub,
    HFSSPINetProcessingSettingsServiceStub,
    HFSSPIPowerGroundNetsServiceStub,
    HFSSPISignalNetsSettingsServiceStub,
)
from ansys.api.edb.v1.hfss_simulation_settings_pb2_grpc import (
    DCRSettingsServiceStub,
    HFSSAdvancedMeshingSettingsServiceStub,
    HFSSAdvancedSettingsServiceStub,
    HFSSGeneralSettingsServiceStub,
    HFSSOptionsSettingsServiceStub,
    HFSSSolverSettingsServiceStub,
)
from ansys.api.edb.v1.hfss_simulation_setup_pb2_grpc import HfssSimulationSetupServiceStub
from ansys.api.edb.v1.hierarchy_obj_pb2_grpc import HierarchyObjectServiceStub
from ansys.api.edb.v1.ic_component_property_pb2_grpc import ICComponentPropertyServiceStub
from ansys.api.edb.v1.inst_array_pb2_grpc import InstArrayServiceStub
from ansys.api.edb.v1.io_component_property_pb2_grpc import IOComponentPropertyServiceStub
from ansys.api.edb.v1.io_manager_pb2_grpc import IOManagerServiceStub
from ansys.api.edb.v1.layer_collection_pb2_grpc import LayerCollectionServiceStub
from ansys.api.edb.v1.layer_map_pb2_grpc import LayerMapServiceStub
from ansys.api.edb.v1.layer_pb2_grpc import LayerServiceStub
from ansys.api.edb.v1.layout_component_pb2_grpc import LayoutComponentServiceStub
from ansys.api.edb.v1.layout_instance_context_pb2_grpc import LayoutInstanceContextServiceStub
from ansys.api.edb.v1.layout_instance_pb2_grpc import LayoutInstanceServiceStub
from ansys.api.edb.v1.layout_obj_instance_2d_geometry_pb2_grpc import (
    LayoutObjInstance2DGeometryServiceStub,
)
from ansys.api.edb.v1.layout_obj_instance_3d_geometry_pb2_grpc import (
    LayoutObjInstance3DGeometryServiceStub,
)
from ansys.api.edb.v1.layout_obj_instance_geometry_pb2_grpc import (
    LayoutObjInstanceGeometryServiceStub,
)
from ansys.api.edb.v1.layout_obj_instance_pb2_grpc import LayoutObjInstanceServiceStub
from ansys.api.edb.v1.layout_obj_pb2_grpc import LayoutObjServiceStub
from ansys.api.edb.v1.layout_pb2_grpc import LayoutServiceStub
from ansys.api.edb.v1.material_def_pb2_grpc import MaterialDefServiceStub
from ansys.api.edb.v1.material_property_thermal_modifier_pb2_grpc import (
    MaterialPropertyThermalModifierServiceStub,
)
from ansys.api.edb.v1.mcad_model_pb2_grpc import McadModelServiceStub
from ansys.api.edb.v1.model_pb2_grpc import ModelServiceStub
from ansys.api.edb.v1.multipole_debye_model_pb2_grpc import MultipoleDebyeModelServiceStub
from ansys.api.edb.v1.net_class_pb2_grpc import NetClassServiceStub
from ansys.api.edb.v1.net_pb2_grpc import NetServiceStub
from ansys.api.edb.v1.netlist_model_pb2_grpc import NetlistModelServiceStub
from ansys.api.edb.v1.package_def_pb2_grpc import PackageDefServiceStub
from ansys.api.edb.v1.padstack_def_data_pb2_grpc import PadstackDefDataServiceStub
from ansys.api.edb.v1.padstack_def_pb2_grpc import PadstackDefServiceStub
from ansys.api.edb.v1.padstack_inst_term_pb2_grpc import PadstackInstanceTerminalServiceStub
from ansys.api.edb.v1.padstack_instance_pb2_grpc import PadstackInstanceServiceStub
from ansys.api.edb.v1.path_pb2_grpc import PathServiceStub
from ansys.api.edb.v1.pin_group_pb2_grpc import PinGroupServiceStub
from ansys.api.edb.v1.pin_group_term_pb2_grpc import PinGroupTerminalServiceStub
from ansys.api.edb.v1.pin_pair_model_pb2_grpc import PinPairModelServiceStub
from ansys.api.edb.v1.point_data_pb2_grpc import PointDataServiceStub
from ansys.api.edb.v1.point_term_pb2_grpc import PointTerminalServiceStub
from ansys.api.edb.v1.polygon_data_pb2_grpc import PolygonDataServiceStub
from ansys.api.edb.v1.polygon_pb2_grpc import PolygonServiceStub
from ansys.api.edb.v1.port_property_pb2_grpc import PortPropertyServiceStub
from ansys.api.edb.v1.primitive_instance_collection_pb2_grpc import (
    PrimitiveInstanceCollectionServiceStub,
)
from ansys.api.edb.v1.primitive_pb2_grpc import PrimitiveServiceStub
from ansys.api.edb.v1.q3d_simulation_settings_pb2_grpc import (
    Q3DAdvancedMeshingSettingsServiceStub,
    Q3DAdvancedSettingsServiceStub,
    Q3DCGSettingsServiceStub,
    Q3DDCRLSettingsServiceStub,
    Q3DGeneralSettingsServiceStub,
    Q3DSettingsServiceStub,
)
from ansys.api.edb.v1.r_tree_pb2_grpc import RTreeServiceStub
from ansys.api.edb.v1.raptor_x_simulation_settings_pb2_grpc import (
    RaptorXAdvancedSettingsServiceStub,
    RaptorXGeneralSettingsServiceStub,
)
from ansys.api.edb.v1.rectangle_pb2_grpc import RectangleServiceStub
from ansys.api.edb.v1.rlc_component_property_pb2_grpc import RLCComponentPropertyServiceStub
from ansys.api.edb.v1.s_parameter_model_pb2_grpc import SParameterModelServiceStub
from ansys.api.edb.v1.si_wave_dcir_simulation_settings_pb2_grpc import (
    SIWaveDCIRSimulationSettingsServiceStub,
)
from ansys.api.edb.v1.si_wave_simulation_settings_pb2_grpc import (
    SIWaveAdvancedSettingsServiceStub,
    SIWaveDCAdvancedSettingsServiceStub,
    SIWaveDCSettingsServiceStub,
    SIWaveGeneralSettingsServiceStub,
    SIWaveSParameterSettingsServiceStub,
)
from ansys.api.edb.v1.simulation_settings_pb2_grpc import (
    AdvancedMeshingSettingsServiceStub,
    AdvancedSettingsServiceStub,
    SettingsOptionsServiceStub,
    SimulationSettingsServiceStub,
    SolverSettingsServiceStub,
)
from ansys.api.edb.v1.simulation_setup_pb2_grpc import SimulationSetupServiceStub
from ansys.api.edb.v1.solder_ball_property_pb2_grpc import SolderBallPropertyServiceStub
from ansys.api.edb.v1.spice_model_pb2_grpc import SpiceModelServiceStub
from ansys.api.edb.v1.stackup_layer_pb2_grpc import StackupLayerServiceStub
from ansys.api.edb.v1.structure_3d_pb2_grpc import Structure3DServiceStub
from ansys.api.edb.v1.term_inst_pb2_grpc import TerminalInstanceServiceStub
from ansys.api.edb.v1.term_inst_term_pb2_grpc import TerminalInstanceTerminalServiceStub
from ansys.api.edb.v1.term_pb2_grpc import TerminalServiceStub
from ansys.api.edb.v1.text_pb2_grpc import TextServiceStub
from ansys.api.edb.v1.transform_3d_pb2_grpc import Transform3DServiceStub
from ansys.api.edb.v1.transform_pb2_grpc import TransformServiceStub
from ansys.api.edb.v1.value_pb2_grpc import ValueServiceStub
from ansys.api.edb.v1.variable_server_pb2_grpc import VariableServerServiceStub
from ansys.api.edb.v1.via_group_pb2_grpc import ViaGroupServiceStub
from ansys.api.edb.v1.via_layer_pb2_grpc import ViaLayerServiceStub
from ansys.api.edb.v1.voltage_regulator_pb2_grpc import VoltageRegulatorServiceStub
import grpc

from ansys.edb.core.inner import LOGGER
from ansys.edb.core.inner.exceptions import EDBSessionException, ErrorCode
from ansys.edb.core.inner.interceptors import ExceptionInterceptor, IOInterceptor

DEFAULT_ADDRESS = "localhost"

# The session module singleton
MOD = modules[__name__]
MOD.current_session = None


class StubAccessor(object):
    """Provides a descriptor for assignig a specific stub to a model."""

    def __init__(self, stub_type: StubType):
        """Initialize a descriptor stub with a name and stub service.

        Parameters
        ----------
        stub_type : .StubType
        """
        self.__stub_name = stub_type.name

    def __get__(self, instance=None, owner=None):
        """Get the corresponding stub service if a session is active."""
        if MOD.current_session is not None:
            return MOD.current_session.stub(self.__stub_name)
        raise EDBSessionException(ErrorCode.NO_SESSIONS)


# Helper class for storing data used by the session
class _Session:
    def __init__(self, ip_address: str, port_num: int, ansys_em_root: str, dump_traffic_log: bool):
        if MOD.current_session is not None:
            raise EDBSessionException(ErrorCode.STARTUP_MULTI_SESSIONS)

        self.ip_address = ip_address or DEFAULT_ADDRESS
        self.port_num = port_num or _find_available_port()
        self.ansys_em_root = ansys_em_root
        self.channel = None
        self.local_server_proc = None
        self.stubs = None
        self.session = None
        self.rpc_counter = defaultdict(int) if dump_traffic_log else None
        self.interceptors = [
            # on reversed order of interception
            IOInterceptor(LOGGER, self.rpc_counter),
            ExceptionInterceptor(LOGGER),
        ]

    def __enter__(self):
        if MOD.current_session == self:
            self.connect()
            return self
        else:
            raise EDBSessionException(ErrorCode.STARTUP_MULTI_SESSIONS)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    def _initialize_stubs(self):
        self.stubs = {stub.name: stub.value(self.channel) for stub in StubType}

    @property
    def server_url(self) -> str:
        return "{}:{}".format(self.ip_address, self.port_num)

    @property
    def server_executable(self) -> str | None:
        if self.is_launch():
            return which(cmd="EDB_RPC_Server", path=self.ansys_em_root)
        else:
            return None

    def server_arguments(self) -> List[str]:
        args = []

        if self.port_num is not None:
            args.append("-p")
            args.append(str(self.port_num))

        return args

    def stub(self, name: str):
        if self.is_active():
            return self.stubs.get(name)

    def is_active(self) -> bool:
        return self.channel is not None and self.stubs is not None

    def is_local(self) -> bool:
        return self.ip_address == "localhost"

    def is_launch(self) -> bool:
        return self.ansys_em_root is not None

    def connect(self):
        if self.is_active():
            return

        if self.is_launch():
            self.start_server()

        self.channel = grpc.insecure_channel(self.server_url)
        self.channel = grpc.intercept_channel(self.channel, *self.interceptors)

        self._initialize_stubs()

    def disconnect(self):
        if self.rpc_counter is not None:
            print(
                "pyedb-core session traffic:" + os.linesep + json.dumps(self.rpc_counter, indent=4)
            )
        if self.stubs is not None:
            self.stubs = None

        if self.channel is not None:
            self.channel.close()
            self.channel = None

        if MOD.current_session == self:
            MOD.current_session = None

        if self.is_launch():
            self.stop_server()

    def start_server(self):
        if not self.is_local():
            return None

        if not self.is_launch():
            return None

        if self.server_executable is None:
            raise EDBSessionException(ErrorCode.STARTUP_NO_EXECUTABLE)

        cmd = [self.server_executable] + self.server_arguments()

        try:
            self.local_server_proc = subprocess.Popen(cmd, stdout=subprocess.PIPE)
        except Exception as e:
            self.disconnect()
            raise EDBSessionException(ErrorCode.STARTUP_UNEXPECTED, e)

        self.wait_server_ready()

    def wait_server_ready(self):
        """Wait for the server to say it successfully started up."""
        local_server_proc_output = self.local_server_proc.stdout.readline()
        stdout = local_server_proc_output.decode().rstrip()

        if not stdout.startswith("Server listening on 127.0.0.1:"):
            try:
                print("Local server failed to start properly. Trying to shut down gracefully...")
                if self.local_server_proc.wait(10):
                    self._on_server_startup_error()
            except subprocess.TimeoutExpired:
                raise EDBSessionException(ErrorCode.STARTUP_TIMEOUT, stdout)
            finally:
                self.disconnect()

    def _on_server_startup_error(self):
        # Check if the server process has exited
        if not self.local_server_proc.poll():
            return

        # Get the return code in a usable format
        ret_code = self.local_server_proc.returncode
        ret_code = ret_code if ret_code < 0 else unpack("i", pack("I", ret_code))[0]

        code = (
            ErrorCode.STARTUP_FAILURE_EDB
            if ret_code == -1
            else ErrorCode.STARTUP_FAILURE_LICENSE
            if ret_code == -2
            else ErrorCode.STARTUP_FAILURE
        )
        raise EDBSessionException(code)

    def stop_server(self):
        if self.local_server_proc and not self.local_server_proc.poll():
            self.local_server_proc.terminate()
            self.local_server_proc.wait()
        self.local_server_proc = None


class StubType(Enum):
    """Provides an enum representing available service stub types."""

    cell = CellServiceStub
    database = DatabaseServiceStub
    layer_collection = LayerCollectionServiceStub
    layer = LayerServiceStub
    stackup_layer = StackupLayerServiceStub
    via_layer = ViaLayerServiceStub
    layout = LayoutServiceStub
    material = MaterialDefServiceStub
    net = NetServiceStub
    primitive = PrimitiveServiceStub
    polygon = PolygonServiceStub
    polygon_data = PolygonDataServiceStub
    path = PathServiceStub
    rectangle = RectangleServiceStub
    via_group = ViaGroupServiceStub
    circle = CircleServiceStub
    text = TextServiceStub
    terminal = TerminalServiceStub
    terminal_instance = TerminalInstanceServiceStub
    terminal_instance_terminal = TerminalInstanceTerminalServiceStub
    bundle_terminal = BundleTerminalServiceStub
    edge = EdgeServiceStub
    edge_terminal = EdgeTerminalServiceStub
    point_terminal = PointTerminalServiceStub
    padstack_instance_terminal = PadstackInstanceTerminalServiceStub
    pin_group = PinGroupServiceStub
    pin_group_terminal = PinGroupTerminalServiceStub
    bondwire = BondwireServiceStub
    bondwire_def = BondwireDefServiceStub
    apd_bondwire_def = ApdBondwireDefServiceStub
    jedec4_bondwire_def = Jedec4BondwireDefServiceStub
    jedec5_bondwire_def = Jedec5BondwireDefServiceStub
    padstack_def = PadstackDefServiceStub
    value = ValueServiceStub
    variable_server = VariableServerServiceStub
    cell_instance = CellInstanceServiceStub
    inst_array = InstArrayServiceStub
    hierarchy_obj = HierarchyObjectServiceStub
    group = GroupServiceStub
    netclass = NetClassServiceStub
    layer_map = LayerMapServiceStub
    point_data = PointDataServiceStub
    arc_data = ArcDataServiceStub
    padstack_instance = PadstackInstanceServiceStub
    voltage_regulator = VoltageRegulatorServiceStub
    connectable = ConnectableServiceStub
    component_group = ComponentGroupServiceStub
    layout_obj = LayoutObjServiceStub
    structure3d = Structure3DServiceStub
    layout_instance = LayoutInstanceServiceStub
    layout_instance_context = LayoutInstanceContextServiceStub
    layout_obj_instance = LayoutObjInstanceServiceStub
    layout_obj_instance_geometry = LayoutObjInstanceGeometryServiceStub
    layout_obj_instance_2d_geometry = LayoutObjInstance2DGeometryServiceStub
    layout_obj_instance_3d_geometry = LayoutObjInstance3DGeometryServiceStub
    component_def = ComponentDefServiceStub
    component_pin = ComponentPinServiceStub
    component_model = ComponentModelServiceStub
    nport_component_model = NPortComponentModelServiceStub
    dyn_link_component_model = DynamicLinkComponentModelServiceStub
    extended_net = ExtendedNetServiceStub
    padstack_def_data = PadstackDefDataServiceStub
    differential_pair = DifferentialPairServiceStub
    solder_ball_property = SolderBallPropertyServiceStub
    component_property = ComponentPropertyServiceStub
    ic_component_property = ICComponentPropertyServiceStub
    die_property = DiePropertyServiceStub
    port_property = PortPropertyServiceStub
    dataset_def = DatasetDefServiceStub
    package_def = PackageDefServiceStub
    dielectric_material_model = DielectricMaterialModelServiceStub
    debye_model = DebyeModelServiceStub
    multipole_debye_model = MultipoleDebyeModelServiceStub
    djordecvic_sarkar_model = DjordjecvicSarkarModelServiceStub
    mcad_model = McadModelServiceStub
    material_property_thermal_modifier = MaterialPropertyThermalModifierServiceStub
    r_tree = RTreeServiceStub
    pin_pair_model = PinPairModelServiceStub
    board_bend_def = BoardBendDefServiceStub
    sparameter_model = SParameterModelServiceStub
    spice_model = SpiceModelServiceStub
    netlist_model = NetlistModelServiceStub
    model = ModelServiceStub
    transform = TransformServiceStub
    io_component_property = IOComponentPropertyServiceStub
    rlc_component_property = RLCComponentPropertyServiceStub
    transform3d = Transform3DServiceStub
    hfss_general_sim_settings = HFSSGeneralSettingsServiceStub
    hfss_options_sim_settings = HFSSOptionsSettingsServiceStub
    hfss_advanced_sim_settings = HFSSAdvancedSettingsServiceStub
    hfss_advanced_sim_meshing_settings = HFSSAdvancedMeshingSettingsServiceStub
    hfss_solver_sim_settings = HFSSSolverSettingsServiceStub
    hfss_dcr_sim_settings = DCRSettingsServiceStub
    hfss_sim_setup = HfssSimulationSetupServiceStub
    hfss_pi_general_sim_settings = HFSSPIGeneralSettingsServiceStub
    hfss_pi_net_processing_sim_settings = HFSSPINetProcessingSettingsServiceStub
    hfss_pi_power_ground_sim_settings = HFSSPIPowerGroundNetsServiceStub
    hfss_pi_signal_nets_sim_settings = HFSSPISignalNetsSettingsServiceStub
    sim_setup = SimulationSetupServiceStub
    sim_settings = SimulationSettingsServiceStub
    sim_settings_options = SettingsOptionsServiceStub
    advanced_sim_settings = AdvancedSettingsServiceStub
    advanced_mesh_sim_settings = AdvancedMeshingSettingsServiceStub
    solver_sim_settings = SolverSettingsServiceStub
    siwave_general_sim_settings = SIWaveGeneralSettingsServiceStub
    siwave_advanced_sim_settings = SIWaveAdvancedSettingsServiceStub
    siwave_dc_sim_settings = SIWaveDCSettingsServiceStub
    siwave_dc_advanced_sim_settings = SIWaveDCAdvancedSettingsServiceStub
    siwave_s_param_sim_settings = SIWaveSParameterSettingsServiceStub
    siwave_dcir_sim_settings = SIWaveDCIRSimulationSettingsServiceStub
    raptor_x_general_sim_settings = RaptorXGeneralSettingsServiceStub
    raptor_x_adv_sim_settings = RaptorXAdvancedSettingsServiceStub
    layout_component = LayoutComponentServiceStub
    io_manager = IOManagerServiceStub
    primitive_instance_collection = PrimitiveInstanceCollectionServiceStub
    edb_error_manager = EDBErrorManagerServiceStub
    q3d_advanced_sim_settings = Q3DAdvancedSettingsServiceStub
    q3d_advanced_meshing_sim_settings = Q3DAdvancedMeshingSettingsServiceStub
    q3d_cg_sim_settings = Q3DCGSettingsServiceStub
    q3d_dcrl_sim_settings = Q3DDCRLSettingsServiceStub
    q3d_general_sim_settings = Q3DGeneralSettingsServiceStub
    q3d_sim_settings = Q3DSettingsServiceStub


def attach_session(
    ip_address: str | None = None, port_num: int = 50051, dump_traffic_log: bool = False
):
    """Attach a session to a port running the EDB API server.

    Parameters
    ----------
    ip_address : str or None, default: None
        IP address of the machine that is running the server. The default is ``None``,
        in which case localhost is used.
    port_num : int, default: 50051
        Port number that the server is listening on.
    dump_traffic_log : bool, default: False
        Flag indicating if the network traffic log should be dumped when the session is disconnected.
    """
    MOD.current_session = _Session(ip_address, port_num, None, dump_traffic_log)
    MOD.current_session.connect()
    return MOD.current_session


def launch_session(ansys_em_root: str, port_num: int | None = None, dump_traffic_log: bool = False):
    r"""Launch a local session to an EDB API server.

    The session must be manually disconnected after use by calling session.disconnect()

    Parameters
    ----------
    ansys_em_root : str
        Directory where the ``EDB_RPC_Server.exe`` file is installed.
    port_num : int or None, default: None
        Port number to listen on. The default is ``None``, in which case a port in [50051, 60000]
        is selected.
    dump_traffic_log : bool, default: False
        Flag indicating if the network traffic log should be dumped when the session is disconnected.

    Examples
    --------
    Create a session and then disconnect it

    >>> session = launch_session("C:\\Program Files\\AnsysEM\\v231\\Win64", 50051)
    >>> # program goes here
    >>> session.disconnect()
    """
    ip_address = None  # remote launch is not supported yet

    try:
        _ensure_session(ansys_em_root, port_num, ip_address, dump_traffic_log)
        return MOD.current_session
    except Exception as e:  # noqa
        if MOD.current_session is not None:
            MOD.current_session.disconnect()
        raise


@contextmanager
def session(
    ansys_em_root: str,
    port_num: int | None = None,
    ip_address: str | None = None,
    dump_traffic_log: bool = False,
):
    r"""Launch a local session to an EDB API server in a context manager.

    Parameters
    ----------
    ansys_em_root : str
        Directory where the ``EDB_RPC_Server.exe`` file is installed.
    port_num : int or None, default: None
        Port number to listen on. The default is ``None``, in which case a port in [50051, 60000]
        is selected.
    ip_address : str or None, default: None
        IP address where the server executable file is running. The default is ``None``, in which
        case localhost is used.

        .. note::
           This parameter is currently not supported. In future releases, this parameter is to
           support remotely running the API on another machine.
    dump_traffic_log : bool, default: False
        Flag indicating if the network traffic log should be dumped when the session is disconnected

    Examples
    --------
    Create a session that automatically disconnects when it goes out of scope.

    >>> with session("C:\\Program Files\\AnsysEM\\v231\\Win64", 50051):
    >>>    # program goes here
    """
    try:
        _ensure_session(ansys_em_root, port_num, ip_address, dump_traffic_log)
        yield
    except EDBSessionException:
        raise
    except Exception as e:  # noqa
        raise
    finally:
        MOD.current_session.disconnect()


def get_layer_collection_stub() -> LayerCollectionServiceStub:
    """Get the layer collection stub.

    Returns
    -------
    .LayerCollectionServiceStub
    """
    return StubAccessor(StubType.layer_collection).__get__()


def get_stackup_layer_stub() -> StackupLayerServiceStub:
    """Get the stackup layer stub.

    Returns
    -------
    .StackupLayerServiceStub
    """
    return StubAccessor(StubType.stackup_layer).__get__()


def get_via_layer_stub() -> ViaLayerServiceStub:
    """Get the via layer stub.

    Returns
    -------
    .ViaLayerServiceStub
    """
    return StubAccessor(StubType.via_layer).__get__()


def get_variable_server_stub() -> VariableServerServiceStub:
    """Get the variable server stub.

    Returns
    -------
    .VariableServerServiceStub
    """
    return StubAccessor(StubType.variable_server).__get__()


def _ensure_session(
    ansys_em_root: str, port_num: int, ip_address: str | None, dump_traffic_log: bool
):
    """Check for a running local session and create one if it doesn't exist.

    Parameters
    ----------
    ansys_em_root : str
        Directory where the ``EDB_RPC_Server.exe`` file is installed.
    port_num : int
        Port number to listen on.
    ip_address : str or None
        IP address where the server executable file is running.
    dump_traffic_log : bool
        Flag indicating if the network traffic log should be dumped when the session is disconnected.
    """
    if MOD.current_session is not None:
        if MOD.current_session.port_num != port_num:
            raise EDBSessionException(ErrorCode.STARTUP_MULTI_SESSIONS)
    else:
        MOD.current_session = _Session(ip_address, port_num, ansys_em_root, dump_traffic_log)
        MOD.current_session.connect()


def _find_available_port(
    interface: str = None, start_port: int = 50051, end_port: int = 60000
) -> int:
    """Find an available port in the given range.

    Parameters
    ----------
    interface : str, default: :data:`DEFAULT_ADDRESS`
        Interface to check for available ports.
    start_port : int, default: ``50051``
        First port number to check.
    end_port : int, default: ``60000``
        Last port number to check.

    Returns
    -------
    int
        First available port in the range.
    """
    for port in range(start_port, end_port):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            if sock.connect_ex((interface or DEFAULT_ADDRESS, port)) == errno.ECONNREFUSED:
                return port
    raise RuntimeError("No available ports found")



"""HFSS simulation setup."""

import ansys.api.edb.v1.hfss_simulation_setup_pb2 as pb

from ansys.edb.core.inner import messages, parser
from ansys.edb.core.inner.utils import map_list
from ansys.edb.core.session import HfssSimulationSetupServiceStub, StubAccessor, StubType
from ansys.edb.core.simulation_setup.hfss_simulation_settings import HFSSSimulationSettings
from ansys.edb.core.simulation_setup.simulation_setup import SimulationSetup, SimulationSetupType


class HfssSimulationSetup(SimulationSetup):
    """Represents HFSS simulation setup data."""

    __stub: HfssSimulationSetupServiceStub = StubAccessor(StubType.hfss_sim_setup)

    @classmethod
    def create(cls, cell, name):
        """Create an HFSS simulation setup.

        Parameters
        ----------
        cell : :class:`.Cell`
            Cell to create the simulation setup in.
        name : str
            Name of the simulation setup.

        Returns
        -------
        HfssSimulationSetup
            HFSS simulation setup created.
        """
        return super()._create(cell, name, SimulationSetupType.HFSS)

    @property
    def mesh_operations(self):
        """:obj:`list` of :class:`.MeshOperation`: Mesh operations of the HFSS simulation setup."""
        return map_list(self.__stub.GetMeshOperations(self.msg).mesh_ops, parser.to_mesh_op)

    @mesh_operations.setter
    def mesh_operations(self, new_mesh_ops):
        self.__stub.SetMeshOperations(
            pb.MeshOperationsPropertyMessage(
                target=self.msg,
                mesh_ops=pb.MeshOperationsMessage(
                    mesh_ops=map_list(new_mesh_ops, messages.mesh_operation_message)
                ),
            )
        )

    @property
    def settings(self):
        """:class:`.HfssSimulationSettings`: Simulation settings of the HFSS simulation setup."""
        return HFSSSimulationSettings(self)
